@@ -10,15 +10,21 @@ l = logging.getLogger(name=__name__)
 
 
 class AMD64ElfGotResolver(IndirectJumpResolver):
+    """
+    A timeless indirect jump resolver that resolves GOT entries on AMD64 ELF binaries.
+    """
+
     def __init__(self, project):
         super().__init__(project, timeless=True)
 
     def filter(self, cfg, addr, func_addr, block, jumpkind):
-        if jumpkind != "Ijk_Call":
+        if not (jumpkind == "Ijk_Call" or (jumpkind == "Ijk_Boring" and addr == func_addr)):
             return False
         return True
 
-    def resolve(self, cfg, addr, func_addr, block, jumpkind):
+    def resolve(  # pylint:disable=unused-argument
+        self, cfg, addr, func_addr, block, jumpkind, func_graph_complete: bool = True, **kwargs
+    ):
         # Find the address and size of the last instruction
         last_insn_addr = None
         last_insn_size = None
@@ -46,6 +52,9 @@ class AMD64ElfGotResolver(IndirectJumpResolver):
             l.warning("Address %# is not mapped.", slot)
             return False, []
 
+        if self.project.loader.find_symbol(target):
+            return True, [target]
+
         if not self.project.is_hooked(target):
             return False, []
 
