@@ -3,16 +3,18 @@ from typing import Generator, Dict, Any, Optional, Set, List
 import operator
 import logging
 
-import networkx
-
-import claripy
 import ailment
+import claripy
+import networkx
+from unique_log_filter import UniqueLogFilter
 
+
+from angr.utils.graph import GraphUtils
 from ...utils.lazy_import import lazy_import
 from ...utils import is_pyinstaller
 from ...utils.graph import dominates, inverted_idoms
 from ...block import Block, BlockNode
-from ..cfg.cfg_utils import CFGUtils
+from .peephole_optimizations import InvertNegatedLogicalConjunctionsAndDisjunctions
 from .structuring.structurer_nodes import (
     MultiNode,
     EmptyBlockNotice,
@@ -28,7 +30,7 @@ from .structuring.structurer_nodes import (
     IncompleteSwitchCaseNode,
 )
 from .graph_region import GraphRegion
-from .utils import first_nonlabel_statement
+from .utils import first_nonlabel_statement, peephole_optimize_expr
 
 if is_pyinstaller():
     # PyInstaller is not happy with lazy import
@@ -38,6 +40,7 @@ else:
 
 
 l = logging.getLogger(__name__)
+l.addFilter(UniqueLogFilter())
 
 
 _UNIFIABLE_COMPARISONS = {
@@ -59,17 +62,23 @@ def _op_with_unified_size(op, conv, operand0, operand1):
     # ensure operand1 is of the same size as operand0
     if isinstance(operand1, ailment.Expr.Const):
         # amazing - we do the eazy thing here
-        return op(conv(operand0), operand1.value)
+        return op(conv(operand0, nobool=True), operand1.value)
     if operand1.bits == operand0.bits:
-        return op(conv(operand0), conv(operand1))
+        return op(conv(operand0, nobool=True), conv(operand1))
     # extension is required
     assert operand1.bits < operand0.bits
     operand1 = ailment.Expr.Convert(None, operand1.bits, operand0.bits, False, operand1)
-    return op(conv(operand0), conv(operand1))
+    return op(conv(operand0, nobool=True), conv(operand1, nobool=True))
 
 
-def _dummy_bvs(condition, condition_mapping):
-    var = claripy.BVS("ailexpr_%s" % repr(condition), condition.bits, explicit_name=True)
+def _dummy_bvs(condition, condition_mapping, name_suffix=""):
+    var = claripy.BVS(f"ailexpr_{repr(condition)}{name_suffix}", condition.bits, explicit_name=True)
+    condition_mapping[var.args[0]] = condition
+    return var
+
+
+def _dummy_bools(condition, condition_mapping, name_suffix=""):
+    var = claripy.BoolS(f"ailexpr_{repr(condition)}{name_suffix}", explicit_name=True)
     condition_mapping[var.args[0]] = condition
     return var
 
@@ -87,23 +96,34 @@ _ail2claripy_op_mapping = {
     "CmpGEs": lambda expr, conv, _: claripy.SGE(conv(expr.operands[0]), conv(expr.operands[1])),
     "CmpGT": lambda expr, conv, _: conv(expr.operands[0]) > conv(expr.operands[1]),
     "CmpGTs": lambda expr, conv, _: claripy.SGT(conv(expr.operands[0]), conv(expr.operands[1])),
-    "Add": lambda expr, conv, _: conv(expr.operands[0]) + conv(expr.operands[1]),
-    "Sub": lambda expr, conv, _: conv(expr.operands[0]) - conv(expr.operands[1]),
-    "Mul": lambda expr, conv, _: conv(expr.operands[0]) * conv(expr.operands[1]),
-    "Div": lambda expr, conv, _: conv(expr.operands[0]) / conv(expr.operands[1]),
+    "Add": lambda expr, conv, _: conv(expr.operands[0], nobool=True) + conv(expr.operands[1], nobool=True),
+    "Sub": lambda expr, conv, _: conv(expr.operands[0], nobool=True) - conv(expr.operands[1], nobool=True),
+    "Mul": lambda expr, conv, _: conv(expr.operands[0], nobool=True) * conv(expr.operands[1], nobool=True),
+    "Div": lambda expr, conv, _: conv(expr.operands[0], nobool=True) / conv(expr.operands[1], nobool=True),
+    "Mod": lambda expr, conv, _: conv(expr.operands[0], nobool=True) % conv(expr.operands[1], nobool=True),
     "Not": lambda expr, conv, _: claripy.Not(conv(expr.operand)),
-    "Xor": lambda expr, conv, _: conv(expr.operands[0]) ^ conv(expr.operands[1]),
-    "And": lambda expr, conv, _: conv(expr.operands[0]) & conv(expr.operands[1]),
-    "Or": lambda expr, conv, _: conv(expr.operands[0]) | conv(expr.operands[1]),
+    "Neg": lambda expr, conv, _: -conv(expr.operand),
+    "BitwiseNeg": lambda expr, conv, _: ~conv(expr.operand),
+    "Xor": lambda expr, conv, _: conv(expr.operands[0], nobool=True) ^ conv(expr.operands[1], nobool=True),
+    "And": lambda expr, conv, _: conv(expr.operands[0], nobool=True) & conv(expr.operands[1], nobool=True),
+    "Or": lambda expr, conv, _: conv(expr.operands[0], nobool=True) | conv(expr.operands[1], nobool=True),
     "Shr": lambda expr, conv, _: _op_with_unified_size(claripy.LShR, conv, expr.operands[0], expr.operands[1]),
     "Shl": lambda expr, conv, _: _op_with_unified_size(operator.lshift, conv, expr.operands[0], expr.operands[1]),
     "Sar": lambda expr, conv, _: _op_with_unified_size(operator.rshift, conv, expr.operands[0], expr.operands[1]),
+    "Concat": lambda expr, conv, _: claripy.Concat(*[conv(operand) for operand in expr.operands]),
     # There are no corresponding claripy operations for the following operations
     "DivMod": lambda expr, _, m: _dummy_bvs(expr, m),
     "CmpF": lambda expr, _, m: _dummy_bvs(expr, m),
     "Mull": lambda expr, _, m: _dummy_bvs(expr, m),
     "Mulls": lambda expr, _, m: _dummy_bvs(expr, m),
     "Reinterpret": lambda expr, _, m: _dummy_bvs(expr, m),
+    "Rol": lambda expr, _, m: _dummy_bvs(expr, m),
+    "Ror": lambda expr, _, m: _dummy_bvs(expr, m),
+    "LogicalXor": lambda expr, _, m: _dummy_bvs(expr, m),
+    "Carry": lambda expr, _, m: _dummy_bvs(expr, m),
+    "SCarry": lambda expr, _, m: _dummy_bvs(expr, m),
+    "SBorrow": lambda expr, _, m: _dummy_bvs(expr, m),
+    "ExpCmpNE": lambda expr, _, m: _dummy_bools(expr, m),
 }
 
 #
@@ -125,6 +145,10 @@ class ConditionProcessor:
         self.guarding_conditions = {}
         self._ast2annotations = {}
 
+        self._peephole_expr_optimizations = [
+            cls(None, None, None) for cls in [InvertNegatedLogicalConjunctionsAndDisjunctions]
+        ]
+
     def clear(self):
         self._condition_mapping = {}
         self.jump_table_conds = defaultdict(set)
@@ -186,7 +210,7 @@ class ConditionProcessor:
 
         reaching_conditions = {}
         # recover the reaching condition for each node
-        sorted_nodes = CFGUtils.quasi_topological_sort_nodes(_g)
+        sorted_nodes = GraphUtils.quasi_topological_sort_nodes(_g)
         terminating_nodes = []
         for node in sorted_nodes:
             # create special conditions for all nodes that are jump table entries
@@ -292,9 +316,11 @@ class ConditionProcessor:
         elif isinstance(node, CodeNode):
             node = CodeNode(
                 self.remove_claripy_bool_asts(node.node, memo=memo),
-                None
-                if node.reaching_condition is None
-                else self.convert_claripy_bool_ast(node.reaching_condition, memo=memo),
+                (
+                    None
+                    if node.reaching_condition is None
+                    else self.convert_claripy_bool_ast(node.reaching_condition, memo=memo)
+                ),
             )
             return node
 
@@ -308,9 +334,11 @@ class ConditionProcessor:
         elif isinstance(node, ConditionNode):
             return ConditionNode(
                 node.addr,
-                None
-                if node.reaching_condition is None
-                else self.convert_claripy_bool_ast(node.reaching_condition, memo=memo),
+                (
+                    None
+                    if node.reaching_condition is None
+                    else self.convert_claripy_bool_ast(node.reaching_condition, memo=memo)
+                ),
                 self.convert_claripy_bool_ast(node.condition, memo=memo),
                 self.remove_claripy_bool_asts(node.true_node, memo=memo),
                 self.remove_claripy_bool_asts(node.false_node, memo=memo),
@@ -461,6 +489,8 @@ class ConditionProcessor:
                     continue
             raise EmptyBlockNotice()
         if type(block) is LoopNode:
+            if block.sequence_node is None:
+                raise EmptyBlockNotice()
             return cls.get_last_statements(block.sequence_node)
         if type(block) is ConditionalBreakNode:
             return [block]
@@ -608,6 +638,8 @@ class ConditionProcessor:
         if cond._hash in memo:
             return memo[cond._hash]
         r = self.convert_claripy_bool_ast_core(cond, memo)
+        optimized_r = peephole_optimize_expr(r, self._peephole_expr_optimizations)
+        r = r if optimized_r is None else optimized_r
         memo[cond._hash] = r
         return r
 
@@ -640,7 +672,8 @@ class ConditionProcessor:
 
         _mapping = {
             "Not": lambda cond_, tags: _unary_op_reduce("Not", cond_.args[0], tags),
-            "Neg": lambda cond_, tags: _unary_op_reduce("Neg", cond_.args[0], tags),
+            "__neg__": lambda cond_, tags: _unary_op_reduce("Not", cond_.args[0], tags),
+            "__invert__": lambda cond_, tags: _unary_op_reduce("BitwiseNeg", cond_.args[0], tags),
             "And": lambda cond_, tags: _binary_op_reduce("LogicalAnd", cond_.args, tags),
             "Or": lambda cond_, tags: _binary_op_reduce("LogicalOr", cond_.args, tags),
             "__le__": lambda cond_, tags: _binary_op_reduce("CmpLE", cond_.args, tags, signed=True),
@@ -666,12 +699,18 @@ class ConditionProcessor:
             "__lshift__": lambda cond_, tags: _binary_op_reduce("Shl", cond_.args, tags),
             "__rshift__": lambda cond_, tags: _binary_op_reduce("Sar", cond_.args, tags),
             "__floordiv__": lambda cond_, tags: _binary_op_reduce("Div", cond_.args, tags),
+            "__mod__": lambda cond_, tags: _binary_op_reduce("Mod", cond_.args, tags),
             "LShR": lambda cond_, tags: _binary_op_reduce("Shr", cond_.args, tags),
             "BVV": lambda cond_, tags: ailment.Expr.Const(None, None, cond_.args[0], cond_.size(), **tags),
-            "BoolV": lambda cond_, tags: ailment.Expr.Const(None, None, True, 1, **tags)
-            if cond_.args[0] is True
-            else ailment.Expr.Const(None, None, False, 1, **tags),
+            "BoolV": lambda cond_, tags: (
+                ailment.Expr.Const(None, None, True, 1, **tags)
+                if cond_.args[0] is True
+                else ailment.Expr.Const(None, None, False, 1, **tags)
+            ),
             "Extract": lambda cond_, tags: self._convert_extract(*cond_.args, tags, memo=memo),
+            "ZeroExt": lambda cond_, tags: _binary_op_reduce(
+                "Concat", [claripy.BVV(0, cond_.args[0]), cond_.args[1]], tags
+            ),
         }
 
         if cond.op in _mapping:
@@ -686,16 +725,18 @@ class ConditionProcessor:
             ("Condition variable %s has an unsupported operator %s. Consider implementing.") % (cond, cond.op)
         )
 
-    def claripy_ast_from_ail_condition(self, condition) -> claripy.ast.Bool:
+    def claripy_ast_from_ail_condition(self, condition, nobool: bool = False) -> claripy.ast.Bool:
         # Unpack a condition all the way to the leaves
         if isinstance(condition, claripy.ast.Base):  # pylint:disable=isinstance-second-argument-not-valid-type
             return condition
 
         if isinstance(
             condition,
-            (ailment.Expr.DirtyExpression, ailment.Expr.BasePointerOffset, ailment.Expr.ITE, ailment.Stmt.Call),
+            (ailment.Expr.DirtyExpression, ailment.Expr.BasePointerOffset, ailment.Expr.ITE),
         ):
             return _dummy_bvs(condition, self._condition_mapping)
+        elif isinstance(condition, ailment.Stmt.Call):
+            return _dummy_bvs(condition, self._condition_mapping, name_suffix=hex(condition.tags.get("ins_addr", 0)))
         elif isinstance(condition, (ailment.Expr.Load, ailment.Expr.Register)):
             # does it have a variable associated?
             if condition.variable is not None:
@@ -709,7 +750,7 @@ class ConditionProcessor:
             self._condition_mapping[var.args[0]] = condition
             return var
         elif isinstance(condition, ailment.Expr.Convert):
-            # convert is special. if it generates a 1-bit variable, it should be treated as a BVS
+            # convert is special. if it generates a 1-bit variable, it should be treated as a BoolS
             if condition.to_bits == 1:
                 var_ = self.claripy_ast_from_ail_condition(condition.operands[0])
                 name = "ailcond_Conv(%d->%d, %d)" % (condition.from_bits, condition.to_bits, hash(var_))
@@ -721,10 +762,12 @@ class ConditionProcessor:
             self._condition_mapping[var.args[0]] = condition
             return var
         elif isinstance(condition, ailment.Expr.Const):
-            if condition.value in {True, False}:
+            if condition.value is True or condition.value is False:
                 var = claripy.BoolV(condition.value)
             else:
                 var = claripy.BVV(condition.value, condition.bits)
+            if isinstance(var, claripy.Bits) and var.size() == 1:
+                var = claripy.true if var.concrete_value == 1 else claripy.false
             return var
         elif isinstance(condition, ailment.Expr.Tmp):
             l.warning("Left-over ailment.Tmp variable %s.", condition)
@@ -753,6 +796,11 @@ class ConditionProcessor:
                 % (condition.op, condition.verbose_op)
             )
         r = lambda_expr(condition, self.claripy_ast_from_ail_condition, self._condition_mapping)
+
+        if isinstance(r, claripy.ast.Bool) and nobool:
+            r = claripy.BVS("ailexpr_from_bool_%r" % r, 1, explicit_name=True)
+            self._condition_mapping[r.args[0]] = condition
+
         if r is NotImplemented:
             if condition.bits == 1:
                 r = claripy.BoolS("ailexpr_%r" % condition, explicit_name=True)
