@@ -53,6 +53,7 @@ class StructurerBase(Analysis):
         case_entry_to_switch_head: Optional[Dict[int, int]] = None,
         parent_region=None,
         improve_structurer=True,
+        **kwargs,
     ):
         self._region: "GraphRegion" = region
         self._parent_map = parent_map
@@ -68,6 +69,9 @@ class StructurerBase(Analysis):
         # intermediate states
         self._new_sequences = []
 
+        # store all virtualized edges (edges that are removed and replaced with a goto)
+        self.virtualized_edges = set()
+
         self.result = None
 
     def _analyze(self):
@@ -141,6 +145,9 @@ class StructurerBase(Analysis):
                 if isinstance(stmt, ailment.Stmt.Jump):
                     targets = extract_jump_targets(stmt)
                     for t in targets:
+                        if t in cases or default is not None and t == default.addr:
+                            # the node after switch cannot be one of the nodes in the switch-case construct
+                            continue
                         goto_addrs[t] += 1
 
         if switch_end_addr is None:
@@ -150,6 +157,8 @@ class StructurerBase(Analysis):
             walker = SequenceWalker(handlers=handlers)
             for case_node in cases.values():
                 walker.walk(case_node)
+            if default is not None:
+                walker.walk(default)
 
             if not goto_addrs:
                 # there is no Goto statement - perfect
@@ -171,8 +180,26 @@ class StructurerBase(Analysis):
                         # remove the last statement
                         block.statements = block.statements[:-1]
 
+        def _handle_Loop(node: LoopNode, parent=None, index=0, label=None):
+            # if a node inside this loop node has a goto that goes to the end of the outer switch-case, we will
+            # convert the goto into a break node, and then add a break node at the end of this switch-case.
+            # of course, this only works if all nodes either end with a return or a goto that goes to the end of the
+            # outer switch-case. we detect it first.
+            # TODO: Implement the above logic
+            return walker._handle_Loop(node, parent=parent, index=index, label=label)
+
+        def _handle_SwitchCase(node: SwitchCaseNode, parent=None, index=0, label=None):
+            # if a node inside this switch-case has a goto that goes to the end of the outer switch-case, we will
+            # convert the goto into a break node, and then add a break node at the end of this switch-case.
+            # of course, this only works if all nodes either end with a return or a goto that goes to the end of the
+            # outer switch-case. we detect it first.
+            # TODO: Implement the above logic
+            return walker._handle_SwitchCase(node, parent=parent, index=index, label=label)
+
         handlers = {
             ailment.Block: _rewrite_gotos,
+            LoopNode: _handle_Loop,
+            SwitchCaseNode: _handle_SwitchCase,
         }
 
         walker = SequenceWalker(handlers=handlers)
@@ -263,6 +290,7 @@ class StructurerBase(Analysis):
                                 ailment.Expr.UnaryOp(None, "Not", jump_stmt.condition),
                                 jump_stmt.false_target,
                                 None,
+                                true_target_idx=jump_stmt.false_target_idx,
                                 **jump_stmt.tags,
                             )
                         elif (
@@ -275,6 +303,7 @@ class StructurerBase(Analysis):
                                 jump_stmt.condition,
                                 jump_stmt.true_target,
                                 None,
+                                true_target_idx=jump_stmt.true_target_idx,
                                 **jump_stmt.tags,
                             )
 
@@ -323,6 +352,7 @@ class StructurerBase(Analysis):
                                 ailment.Expr.UnaryOp(None, "Not", jump_stmt.condition),
                                 jump_stmt.false_target,
                                 None,
+                                true_target_idx=jump_stmt.false_target_idx,
                                 **jump_stmt.tags,
                             )
                         elif (
@@ -335,6 +365,7 @@ class StructurerBase(Analysis):
                                 jump_stmt.condition,
                                 jump_stmt.true_target,
                                 None,
+                                true_target_idx=jump_stmt.false_target_idx,
                                 **jump_stmt.tags,
                             )
 
@@ -818,6 +849,7 @@ class StructurerBase(Analysis):
                         ailment.Expr.UnaryOp(None, "Not", last_stmt.condition),
                         last_stmt.false_target,
                         None,
+                        true_target_idx=last_stmt.false_target_idx,
                         **last_stmt.tags,
                     )
                     last_node.statements[-1] = new_stmt
@@ -830,6 +862,7 @@ class StructurerBase(Analysis):
                         last_stmt.condition,
                         last_stmt.true_target,
                         None,
+                        true_target_idx=last_stmt.true_target_idx,
                         **last_stmt.tags,
                     )
                     last_node.statements[-1] = new_stmt
