@@ -1,3 +1,4 @@
+from typing import TYPE_CHECKING
 from collections import defaultdict
 
 import claripy
@@ -9,6 +10,10 @@ from ..storage.memory_mixins.paged_memory.pages.multi_values import MultiValues
 from ..knowledge_plugins.key_definitions.constants import OP_BEFORE, OP_AFTER
 from . import Analysis, VtableFinder, CFGFast, ReachingDefinitionsAnalysis
 
+if TYPE_CHECKING:
+    from angr.analyses.reaching_definitions.function_handler import FunctionCallData
+    from angr.analyses.reaching_definitions.rd_state import ReachingDefinitionsState
+
 
 class PossibleObject:
     """
@@ -40,6 +45,7 @@ class NewFunctionHandler(FunctionHandler):
     """
 
     def __init__(self, max_addr=None, new_func_addr=None, project=None):
+        super().__init__()
         self.max_addr = max_addr
 
         # this is a map between an object addr outside the mapped binary and PossibleObject instance
@@ -54,15 +60,10 @@ class NewFunctionHandler(FunctionHandler):
 
     def handle_local_function(
         self,
-        state,
-        function_address,
-        call_stack,
-        maximum_local_call_depth,
-        visited_blocks,
-        dep_graph,
-        src_ins_addr=None,
-        codeloc=None,
+        state: "ReachingDefinitionsState",
+        data: "FunctionCallData",
     ):
+        function_address = data.address
         if function_address == self.new_func_addr:
             word_size = self.project.arch.bits // self.project.arch.byte_width
             # check if this is a call to new()
@@ -74,8 +75,8 @@ class NewFunctionHandler(FunctionHandler):
             else:
                 size_arg_reg_offset = self.project.arch.registers["rdi"][0]
                 size_arg_reg_size = word_size
-            v0 = state.register_definitions.load(size_arg_reg_offset, size_arg_reg_size).one_value()
-            size = v0._model_concrete.value if v0 is not None else None
+            v0 = state.registers.load(size_arg_reg_offset, size_arg_reg_size).one_value()
+            size = v0.concrete_value if v0 is not None and v0.concrete else None
 
             if size is not None:
                 # None since we do not know it's class yet, it is a possible this pointer
@@ -89,10 +90,11 @@ class NewFunctionHandler(FunctionHandler):
                 ret_val_reg_offset = self.project.arch.registers["rax"][0]
                 ret_val_reg_size = word_size
             atom = Register(ret_val_reg_offset, ret_val_reg_size)
-            state.kill_and_add_definition(
+            data.depends(
                 atom,
-                codeloc,
-                MultiValues(offset_to_values={0: {claripy.BVV(self.max_addr, word_size * state.arch.byte_width)}}),
+                value=MultiValues(
+                    offset_to_values={0: {claripy.BVV(self.max_addr, word_size * state.arch.byte_width)}}
+                ),
             )
             # setting the values pointed by rax to zero
             memory_location = MemoryLocation(self.max_addr, size)
@@ -100,21 +102,23 @@ class NewFunctionHandler(FunctionHandler):
 
             for offset in range(0, size, word_size):
                 offset_to_values[offset] = {claripy.BVV(0, word_size * state.arch.byte_width)}
-            state.kill_and_add_definition(memory_location, codeloc, MultiValues(offset_to_values=offset_to_values))
+            data.depends(memory_location, value=MultiValues(offset_to_values=offset_to_values))
             self.max_addr += size
 
-        elif "ctor" in self.project.kb.functions[function_address].demangled_name:
-            # check if rdi has a possible this pointer/ object address, if so then we can assign this object this class
-            # also if the func is a constructor(not stripped binaries)
-            for addr, possible_object in self.possible_objects_dict.items():
-                v1 = state.register_definitions.load(72, state.arch.bits // state.arch.byte_width).one_value()
-                obj_addr = v1._model_concrete.value if v1 is not None else None
-                if obj_addr is not None and addr == obj_addr:
-                    col_ind = self.project.kb.functions[function_address].demangled_name.rfind("::")
-                    class_name = self.project.kb.functions[function_address].demangled_name[:col_ind]
-                    possible_object.class_name = class_name
-        executed_rda = True
-        return executed_rda, state, visited_blocks, dep_graph
+        else:
+            if self.project.kb.functions.contains_addr(function_address):
+                func = self.project.kb.functions.get_by_addr(function_address)
+                if func is not None and "ctor" in func.demangled_name:
+                    # check if rdi has a possible this pointer/ object address, if so then we can assign this object
+                    # this class
+                    # also if the func is a constructor(not stripped binaries)
+                    for addr, possible_object in self.possible_objects_dict.items():
+                        v1 = state.registers.load(72, state.arch.bits // state.arch.byte_width).one_value()
+                        obj_addr = v1.concrete_value if v1 is not None and v1.concrete else None
+                        if obj_addr is not None and addr == obj_addr:
+                            col_ind = self.project.kb.functions[function_address].demangled_name.rfind("::")
+                            class_name = self.project.kb.functions[function_address].demangled_name[:col_ind]
+                            possible_object.class_name = class_name
 
 
 class StaticObjectFinder(Analysis):
@@ -141,15 +145,18 @@ class StaticObjectFinder(Analysis):
         # this is the addr where all the this pointers returned by new() will be pointing to
         max_addr = self.project.loader.main_object.max_addr + 8
         new_func_addr = self.project.kb.functions["_Znwm"].addr
-        for func in all_functions:
-            cc = self.project.kb.functions[func].calling_convention
+        for func in all_functions.values():
+            if func.is_simprocedure:
+                continue
+
+            cc = func.calling_convention
             word_size = self.project.arch.bits // self.project.arch.byte_width
             # the map fo this ptrs
             newhandler = NewFunctionHandler(max_addr=max_addr, new_func_addr=new_func_addr, project=self.project)
             max_addr = newhandler.max_addr
             # this performs RDA as well as mark possible object instances for non stripped binaries
             rd = self.project.analyses[ReachingDefinitionsAnalysis].prep()(
-                all_functions[func], observe_all=True, function_handler=newhandler
+                func, observe_all=True, function_handler=newhandler
             )
             for addr, pos_obj in newhandler.possible_objects_dict.items():
                 self.possible_objects[addr] = pos_obj
@@ -157,10 +164,11 @@ class StaticObjectFinder(Analysis):
             # for stripped binary we check if the first function called after new(),
             # is passed the this pointer(returned by new)...
             # if so then we say that it is possibly a constructor
-            for node in all_functions[func].graph.nodes():
-                if all_functions[func].get_call_target(node.addr) == new_func_addr:
-                    ret_node_addr = all_functions[func].get_call_return(node.addr)
-                    call_after_new_addr = all_functions[func].get_call_target(ret_node_addr)
+            for node in func.graph.nodes():
+                if func.get_call_target(node.addr) == new_func_addr:
+                    ret_node_addr = func.get_call_return(node.addr)
+                    ret_node = self.project.factory.block(ret_node_addr)
+                    call_after_new_addr = func.get_call_target(ret_node_addr)
                     rd_before_node = rd.get_reaching_definitions_by_node(ret_node_addr, OP_BEFORE)
 
                     if cc is not None:
@@ -169,10 +177,15 @@ class StaticObjectFinder(Analysis):
                     else:
                         ret_val_reg_offset = self.project.arch.registers["rax"][0]
                         ret_val_reg_size = word_size
-                    v0 = rd_before_node.register_definitions.load(ret_val_reg_offset, ret_val_reg_size).one_value()
-                    addr_of_new_obj = v0._model_concrete.value if v0 is not None else None
+                    v0 = rd_before_node.registers.load(ret_val_reg_offset, ret_val_reg_size).one_value()
+                    addr_of_new_obj = v0.concrete_value if v0 is not None and v0.concrete else None
 
-                    rd_after_node = rd.get_reaching_definitions_by_node(ret_node_addr, OP_AFTER)
+                    # we need the state right before the call
+                    if ret_node.vex.jumpkind == "Ijk_Call" and ret_node.vex.instruction_addresses:
+                        last_ins_addr = ret_node.instruction_addrs[-1]
+                        rd_after_node = rd.get_reaching_definitions_by_insn(last_ins_addr, OP_BEFORE)
+                    else:
+                        rd_after_node = rd.get_reaching_definitions_by_insn(ret_node_addr, OP_AFTER)
 
                     if cc is not None:
                         this_ptr_reg_offset = self.project.arch.registers[cc.args[0].reg_name][0]
@@ -180,8 +193,8 @@ class StaticObjectFinder(Analysis):
                     else:
                         this_ptr_reg_offset = self.project.arch.registers["rdi"][0]
                         this_ptr_reg_size = word_size
-                    v1 = rd_after_node.register_definitions.load(this_ptr_reg_offset, this_ptr_reg_size).one_value()
-                    addr_in_rdi = v1._model_concrete.value if v1 is not None else None
+                    v1 = rd_after_node.registers.load(this_ptr_reg_offset, this_ptr_reg_size).one_value()
+                    addr_in_rdi = v1.concrete_value if v1 is not None and v1.concrete else None
 
                     if addr_of_new_obj is not None and addr_of_new_obj == addr_in_rdi:
                         self.possible_constructors[call_after_new_addr].append(self.possible_objects[addr_of_new_obj])
