@@ -1,11 +1,11 @@
 import logging
-from typing import Optional, DefaultDict, Dict, List, Tuple, Set, Any, Union, TYPE_CHECKING
+from typing import Optional, DefaultDict, Dict, Tuple, Set, Any, Union, TYPE_CHECKING, Iterable
 from collections import defaultdict
 
 import ailment
 import pyvex
-from ..forward_analysis.visitors.graph import NodeType
 
+from angr.analyses import ForwardAnalysis
 from ...block import Block
 from ...knowledge_plugins.cfg.cfg_node import CFGNode
 from ...codenode import CodeNode
@@ -13,26 +13,30 @@ from ...engines.light import SimEngineLight
 from ...knowledge_plugins.functions import Function
 from ...knowledge_plugins.key_definitions import ReachingDefinitionsModel, LiveDefinitions
 from ...knowledge_plugins.key_definitions.constants import OP_BEFORE, OP_AFTER, ObservationPointType
+from ...code_location import CodeLocation, ExternalCodeLocation
 from ...misc.ux import deprecated
+from ..forward_analysis.visitors.graph import NodeType
 from ..analysis import Analysis
-from ..forward_analysis import ForwardAnalysis
 from .engine_ail import SimEngineRDAIL
 from .engine_vex import SimEngineRDVEX
 from .rd_state import ReachingDefinitionsState
+from .rd_initializer import RDAStateInitializer
 from .subject import Subject, SubjectType
-from .function_handler import FunctionHandler
+from .function_handler import FunctionHandler, FunctionCallRelationships
+from .dep_graph import DepGraph
 
 if TYPE_CHECKING:
-    from .dep_graph import DepGraph
-    from typing import Literal, Iterable
+    from typing import Literal
 
-    ObservationPoint = Tuple[Literal["insn", "node"], int, ObservationPointType]
+    ObservationPoint = Tuple[
+        Literal["insn", "node", "stmt", "exit"], Union[int, Tuple[int, int], Tuple[int, int, int]], ObservationPointType
+    ]
 
 l = logging.getLogger(name=__name__)
 
 
 class ReachingDefinitionsAnalysis(
-    ForwardAnalysis[ReachingDefinitionsState, NodeType], Analysis
+    ForwardAnalysis[ReachingDefinitionsState, NodeType, object, object], Analysis
 ):  # pylint:disable=abstract-method
     """
     ReachingDefinitionsAnalysis is a text-book implementation of a static data-flow analysis that works on either a
@@ -49,31 +53,35 @@ class ReachingDefinitionsAnalysis(
 
     def __init__(
         self,
-        subject: Union[Subject, ailment.Block, Block, Function] = None,
+        subject: Union[Subject, ailment.Block, Block, Function, str] = None,
         func_graph=None,
-        max_iterations=3,
+        max_iterations=30,
         track_tmps=False,
-        track_calls=None,
-        track_consts=False,
+        track_consts=True,
         observation_points: "Iterable[ObservationPoint]" = None,
         init_state: ReachingDefinitionsState = None,
+        init_context=None,
+        state_initializer: Optional["RDAStateInitializer"] = None,
         cc=None,
         function_handler: "Optional[FunctionHandler]" = None,
-        call_stack: Optional[List[int]] = None,
-        maximum_local_call_depth=5,
         observe_all=False,
         visited_blocks=None,
-        dep_graph: Optional["DepGraph"] = None,
+        dep_graph: Union[DepGraph, bool, None] = True,
         observe_callback=None,
         canonical_size=8,
+        stack_pointer_tracker=None,
+        use_callee_saved_regs_at_return=True,
+        interfunction_level: int = 0,
+        track_liveness: bool = True,
+        func_addr: Optional[int] = None,
+        element_limit: int = 5,
     ):
         """
         :param subject:                         The subject of the analysis: a function, or a single basic block
         :param func_graph:                      Alternative graph for function.graph.
-        :param int max_iterations:              The maximum number of iterations before the analysis is terminated.
-        :param bool track_tmps:                 Whether or not temporary variables should be taken into consideration
+        :param max_iterations:                  The maximum number of iterations before the analysis is terminated.
+        :param track_tmps:                      Whether or not temporary variables should be taken into consideration
                                                 during the analysis.
-        :param bool track_calls:                Whether or not calls will show up as elements in the def-use graph.
         :param iterable observation_points:     A collection of tuples of ("node"|"insn", ins_addr, OP_TYPE) defining
                                                 where reaching definitions should be copied and stored. OP_TYPE can be
                                                 OP_BEFORE or OP_AFTER.
@@ -81,24 +89,31 @@ class ReachingDefinitionsAnalysis(
                                                 copy.
                                                 Default to None: the analysis then initialize its own abstract state,
                                                 based on the given <Subject>.
-        :param SimCC cc:                        Calling convention of the function.
-        :param FunctionHandler function_handler:
-                                                The function handler to update the analysis state and results on
+        :param init_context:                    If init_state is not given, this is used to initialize the context
+                                                field of the initial state's CodeLocation. The only default-supported
+                                                type which may go here is a tuple of integers, i.e. a callstack.
+                                                Anything else requires a custom FunctionHandler.
+        :param cc:                              Calling convention of the function.
+        :param function_handler:                The function handler to update the analysis state and results on
                                                 function calls.
-        :param call_stack:                      An ordered list of Function addresses representing the call stack
-                                                leading to the analysed subject, from older to newer calls. Setting it
-                                                to None to limit the analysis to a single function and disable call
-                                                stack tracking; In that case, all contexts in CodeLocation will be
-                                                None, which makes CodeLocation objects contextless.
-        :param int maximum_local_call_depth:    Maximum local function recursion depth.
-        :param Boolean observe_all:             Observe every statement, both before and after.
+        :param observe_all:                     Observe every statement, both before and after.
         :param visited_blocks:                  A set of previously visited blocks.
         :param dep_graph:                       An initial dependency graph to add the result of the analysis to. Set it
                                                 to None to skip dependency graph generation.
         :param canonical_size:                  The sizes (in bytes) that objects with an UNKNOWN_SIZE are treated as
                                                 for operations where sizes are necessary.
+        :param dep_graph:                       Set this to True to generate a dependency graph for the subject. It will
+                                                be available as `result.dep_graph`.
+        :param interfunction_level:             The number of functions we should recurse into. This parameter is only
+                                                used if function_handler is not provided.
+        :param track_liveness:                  Whether to track liveness information. This can consume
+                                                sizeable amounts of RAM on large functions. (e.g. ~15GB for a function
+                                                with 4k nodes)
+
         """
 
+        if isinstance(subject, str):
+            subject = self.kb.functions[subject]
         if not isinstance(subject, Subject):
             self._subject = Subject(subject, func_graph, cc)
         else:
@@ -110,28 +125,38 @@ class ReachingDefinitionsAnalysis(
         )
 
         self._track_tmps = track_tmps
-        self._track_calls = track_calls
         self._track_consts = track_consts
         self._max_iterations = max_iterations
         self._observation_points = observation_points
         self._init_state = init_state
-        self._maximum_local_call_depth = maximum_local_call_depth
         self._canonical_size = canonical_size
+        self._use_callee_saved_regs_at_return = use_callee_saved_regs_at_return
+        self._func_addr = func_addr
+        self._element_limit = element_limit
 
-        self._dep_graph = dep_graph
+        if dep_graph is None or dep_graph is False:
+            self._dep_graph = None
+        elif dep_graph is True:
+            self._dep_graph = DepGraph()
+        else:
+            self._dep_graph = dep_graph
 
         if function_handler is None:
-            self._function_handler = FunctionHandler().hook(self)
+            self._function_handler = FunctionHandler(interfunction_level).hook(self)
         else:
+            if interfunction_level != 0:
+                l.warning("RDA(interfunction_level=XXX) doesn't do anything if you provide a function handler")
             self._function_handler = function_handler.hook(self)
 
-        self._call_stack: Optional[List[int]] = None
-        if call_stack is not None:
-            self._call_stack = self._init_call_stack(call_stack, subject)
-
         if self._init_state is not None:
             self._init_state = self._init_state.copy()
             self._init_state.analysis = self
+            # There should never be an initializer needed then
+            self._state_initializer = None
+        else:
+            self._state_initializer = state_initializer
+
+        self._init_context = init_context
 
         self._observe_all = observe_all
         self._observe_callback = observe_callback
@@ -142,47 +167,42 @@ class ReachingDefinitionsAnalysis(
 
         self._node_iterations: DefaultDict[int, int] = defaultdict(int)
 
+        self.model: ReachingDefinitionsModel = ReachingDefinitionsModel(
+            func_addr=self.subject.content.addr if isinstance(self.subject.content, Function) else None,
+            track_liveness=track_liveness,
+        )
+
+        bp_as_gpr = False
+        the_func = None
+        if isinstance(self.subject.content, Function):
+            the_func = self.subject.content
+        else:
+            if self._func_addr is not None:
+                try:
+                    the_func = self.kb.functions.get_by_addr(self._func_addr)
+                except KeyError:
+                    pass
+        if the_func is not None:
+            bp_as_gpr = the_func.info.get("bp_as_gpr", False)
+
         self._engine_vex = SimEngineRDVEX(
             self.project,
-            self._call_stack,
-            self._maximum_local_call_depth,
             functions=self.kb.functions,
             function_handler=self._function_handler,
         )
         self._engine_ail = SimEngineRDAIL(
-            self.project, self._call_stack, self._maximum_local_call_depth, function_handler=self._function_handler
+            self.project,
+            function_handler=self._function_handler,
+            stack_pointer_tracker=stack_pointer_tracker,
+            use_callee_saved_regs_at_return=self._use_callee_saved_regs_at_return,
+            bp_as_gpr=bp_as_gpr,
         )
 
         self._visited_blocks: Set[Any] = visited_blocks or set()
-        self.model: ReachingDefinitionsModel = ReachingDefinitionsModel(
-            func_addr=self.subject.content.addr if isinstance(self.subject.content, Function) else None
-        )
+        self.function_calls: Dict[CodeLocation, FunctionCallRelationships] = {}
 
         self._analyze()
 
-    def _init_call_stack(self, call_stack: List[int], subject) -> List[int]:
-        if self._subject.type == SubjectType.Function:
-            return call_stack + [subject.addr]
-        elif self._subject.type == SubjectType.Block:
-            cfg = self.kb.cfgs.get_most_accurate()
-            if cfg is None:
-                # no CFG exists
-                return call_stack
-            cfg_node = cfg.get_any_node(subject.addr)
-            if cfg_node is None:
-                # we don't know which function this node belongs to
-                return call_stack
-            function_address = cfg_node.function_address
-            function = self.kb.functions.function(function_address)
-            if len(call_stack) > 0 and call_stack[-1] == function.addr:
-                return call_stack
-            else:
-                return call_stack + [function.addr]
-        elif self._subject.type == SubjectType.CallTrace:
-            return call_stack + [self._subject.content.current_function_address()]
-        else:
-            raise ValueError("Unexpected subject type %s." % self._subject.type)
-
     @property
     def observed_results(self) -> Dict[Tuple[str, int, int], LiveDefinitions]:
         return self.model.observed_results
@@ -209,16 +229,18 @@ class ReachingDefinitionsAnalysis(
         return next(iter(self.observed_results.values()))
 
     @property
-    def dep_graph(self):
+    def dep_graph(self) -> DepGraph:
+        if self._dep_graph is None:
+            raise ValueError(
+                "Cannot access dep_graph if the analysis was not configured to generate one. Try passing "
+                "dep_graph=True to the RDA constructor."
+            )
         return self._dep_graph
 
     @property
     def visited_blocks(self):
         return self._visited_blocks
 
-    def _current_local_call_depth(self):
-        return len(self._call_stack)
-
     @deprecated(replacement="get_reaching_definitions_by_insn")
     def get_reaching_definitions(self, ins_addr, op_type):
         return self.get_reaching_definitions_by_insn(ins_addr, op_type)
@@ -227,11 +249,8 @@ class ReachingDefinitionsAnalysis(
         key = "insn", ins_addr, op_type
         if key not in self.observed_results:
             raise KeyError(
-                (
-                    "Reaching definitions are not available at observation point %s. "
-                    "Did you specify that observation point?"
-                )
-                % key
+                "Reaching definitions are not available at observation point %s. "
+                "Did you specify that observation point?" % str(key)
             )
 
         return self.observed_results[key]
@@ -246,11 +265,18 @@ class ReachingDefinitionsAnalysis(
 
         return self.observed_results[key]
 
-    def node_observe(self, node_addr: int, state: ReachingDefinitionsState, op_type: ObservationPointType) -> None:
+    def node_observe(
+        self,
+        node_addr: int,
+        state: ReachingDefinitionsState,
+        op_type: ObservationPointType,
+        node_idx: Optional[int] = None,
+    ) -> None:
         """
         :param node_addr:   Address of the node.
         :param state:       The analysis state.
-        :param op_type:     Type of the bbservation point. Must be one of the following: OP_BEFORE, OP_AFTER.
+        :param op_type:     Type of the observation point. Must be one of the following: OP_BEFORE, OP_AFTER.
+        :param node_idx:    ID of the node. Used in AIL to differentiate blocks with the same address.
         """
 
         key = None
@@ -259,15 +285,21 @@ class ReachingDefinitionsAnalysis(
 
         if self._observe_all:
             observe = True
-            key: ObservationPoint = ("node", node_addr, op_type)
+            key: ObservationPoint = (
+                ("node", node_addr, op_type) if node_idx is None else ("node", (node_addr, node_idx), op_type)
+            )
         elif self._observation_points is not None:
-            key: ObservationPoint = ("node", node_addr, op_type)
+            key: ObservationPoint = (
+                ("node", node_addr, op_type) if node_idx is None else ("node", (node_addr, node_idx), op_type)
+            )
             if key in self._observation_points:
                 observe = True
         elif self._observe_callback is not None:
-            observe = self._observe_callback("node", addr=node_addr, state=state, op_type=op_type)
+            observe = self._observe_callback("node", addr=node_addr, state=state, op_type=op_type, node_idx=node_idx)
             if observe:
-                key: ObservationPoint = ("node", node_addr, op_type)
+                key: ObservationPoint = (
+                    ("node", node_addr, op_type) if node_idx is None else ("node", (node_addr, node_idx), op_type)
+                )
 
         if observe:
             self.observed_results[key] = state.live_definitions
@@ -323,6 +355,98 @@ class ReachingDefinitionsAnalysis(
             # it's an AIL block
             self.observed_results[key] = state.live_definitions.copy()
 
+    def stmt_observe(
+        self,
+        stmt_idx: int,
+        stmt: Union[ailment.Stmt.Statement, pyvex.stmt.IRStmt],
+        block: Union[Block, ailment.Block],
+        state: ReachingDefinitionsState,
+        op_type: ObservationPointType,
+    ) -> None:
+        """
+
+        :param stmt_idx:
+        :param stmt:
+        :param block:
+        :param state:
+        :param op_type:
+        :return:
+        """
+
+        key = None
+        observe = False
+
+        block_idx = block.idx if isinstance(block, ailment.Block) else None
+        if self._observe_all:
+            observe = True
+            key: ObservationPoint = ("stmt", (block.addr, block_idx, stmt_idx), op_type)
+        elif self._observation_points is not None:
+            key: ObservationPoint = ("stmt", (block.addr, block_idx, stmt_idx), op_type)
+            if key in self._observation_points:
+                observe = True
+        elif self._observe_callback is not None:
+            observe = self._observe_callback(
+                "stmt", stmt_idx=stmt_idx, stmt=stmt, block=block, state=state, op_type=op_type
+            )
+            if observe:
+                key: ObservationPoint = ("stmt", (block.addr, block_idx, stmt_idx), op_type)
+
+        if not observe:
+            return
+
+        if isinstance(stmt, pyvex.stmt.IRStmt):
+            # it's an angr block
+            self.observed_results[key] = state.live_definitions.copy()
+        elif isinstance(stmt, ailment.Stmt.Statement):
+            # it's an AIL block
+            self.observed_results[key] = state.live_definitions.copy()
+
+    def exit_observe(
+        self,
+        node_addr: int,
+        exit_stmt_idx: int,
+        block: Union[Block, ailment.Block],
+        state: ReachingDefinitionsState,
+        node_idx: Optional[int] = None,
+    ):
+        observe = False
+        key = None
+
+        if self._observe_all:
+            observe = True
+            key = (
+                ("exit", (node_addr, exit_stmt_idx), ObservationPointType.OP_AFTER)
+                if node_idx is None
+                else ("exit", (node_addr, node_idx, exit_stmt_idx), ObservationPointType.OP_AFTER)
+            )
+        elif self._observation_points is not None:
+            key = (
+                ("exit", (node_addr, exit_stmt_idx), ObservationPointType.OP_AFTER)
+                if node_idx is None
+                else ("exit", (node_addr, node_idx, exit_stmt_idx), ObservationPointType.OP_AFTER)
+            )
+            if key in self._observation_points:
+                observe = True
+        elif self._observe_callback is not None:
+            observe = self._observe_callback(
+                "exit",
+                node_addr=node_addr,
+                exit_stmt_idx=exit_stmt_idx,
+                block=block,
+                state=state,
+            )
+            if observe:
+                key = (
+                    ("exit", (node_addr, exit_stmt_idx), ObservationPointType.OP_AFTER)
+                    if node_idx is None
+                    else ("exit", (node_addr, node_idx, exit_stmt_idx), ObservationPointType.OP_AFTER)
+                )
+
+        if not observe:
+            return
+
+        self.observed_results[key] = state.live_definitions.copy()
+
     @property
     def subject(self):
         return self._subject
@@ -334,25 +458,41 @@ class ReachingDefinitionsAnalysis(
     def _pre_analysis(self):
         pass
 
-    def _initial_abstract_state(self, _node) -> ReachingDefinitionsState:
+    def _initial_abstract_state(self, node) -> ReachingDefinitionsState:
         if self._init_state is not None:
             return self._init_state
         else:
             return ReachingDefinitionsState(
+                CodeLocation(node.addr, stmt_idx=0, ins_addr=node.addr, context=self._init_context),
                 self.project.arch,
                 self.subject,
                 track_tmps=self._track_tmps,
-                track_calls=self._track_calls,
                 track_consts=self._track_consts,
                 analysis=self,
                 canonical_size=self._canonical_size,
+                initializer=self._state_initializer,
+                element_limit=self._element_limit,
             )
 
-    # pylint: disable=no-self-use
+    # pylint: disable=no-self-use,arguments-differ
     def _merge_states(self, _node, *states: ReachingDefinitionsState):
+        assert len(states) >= 2
         merged_state, merge_occurred = states[0].merge(*states[1:])
         return merged_state, not merge_occurred
 
+    def _compare_states(self, node, old_state: ReachingDefinitionsState, new_state: ReachingDefinitionsState) -> bool:
+        """
+        Return True if new_state >= old_state in the lattice.
+
+        :param node:
+        :param old_state:
+        :param new_state:
+        :return:
+        """
+
+        reached_fixedpoint = new_state.compare(old_state)
+        return reached_fixedpoint
+
     def _run_on_node(self, node, state: ReachingDefinitionsState):
         """
 
@@ -367,31 +507,46 @@ class ReachingDefinitionsAnalysis(
 
         if isinstance(node, ailment.Block):
             block = node
+            block_key = (node.addr, node.idx)
             engine = self._engine_ail
         elif isinstance(node, (Block, CodeNode)):
             block = self.project.factory.block(node.addr, node.size, opt_level=1, cross_insn_opt=False)
             engine = self._engine_vex
+            block_key = node.addr
         elif isinstance(node, CFGNode):
             if node.is_simprocedure or node.is_syscall:
-                return False, state.copy()
+                return False, state.copy(discard_tmpdefs=True)
             block = node.block
             engine = self._engine_vex
+            block_key = node.addr
         else:
             l.warning("Unsupported node type %s.", node.__class__)
-            return False, state.copy()
+            return False, state.copy(discard_tmpdefs=True)
 
-        self.node_observe(node.addr, state, OP_BEFORE)
+        state = state.copy(discard_tmpdefs=True)
+        self.node_observe(node.addr, state.copy(), OP_BEFORE)
 
-        state = state.copy()
-        state, self._visited_blocks, self._dep_graph = engine.process(
+        if self.subject.type == SubjectType.Function:
+            node_parents = [
+                CodeLocation(pred.addr, 0, block_idx=pred.idx if isinstance(pred, ailment.Block) else None)
+                for pred in self._graph_visitor.predecessors(node)
+            ]
+            if node.addr == self.subject.content.addr:
+                node_parents += [ExternalCodeLocation()]
+            self.model.at_new_block(
+                CodeLocation(block.addr, 0, block_idx=block.idx if isinstance(block, ailment.Block) else None),
+                node_parents,
+            )
+
+        state = engine.process(
             state,
             block=block,
             fail_fast=self._fail_fast,
             visited_blocks=self._visited_blocks,
             dep_graph=self._dep_graph,
+            model=self.model,
         )
 
-        block_key = node.addr
         self._node_iterations[block_key] += 1
 
         self.node_observe(node.addr, state, OP_AFTER)
@@ -401,8 +556,18 @@ class ReachingDefinitionsAnalysis(
         for use in [state.stack_uses, state.heap_uses, state.register_uses, state.memory_uses]:
             self.all_uses.merge(use)
 
+        if self._track_tmps:
+            # merge tmp uses to all_uses
+            for tmp_idx, locs in state.tmp_uses.items():
+                tmp_def = next(iter(state.tmps[tmp_idx]))
+                for loc in locs:
+                    self.all_uses.add_use(tmp_def, loc)
+
+        # drop definitions and uses because we will not need them anymore
+        state.downsize()
+
         if self._node_iterations[block_key] < self._max_iterations:
-            return True, state
+            return None, state
         else:
             return False, state
 
@@ -411,3 +576,18 @@ class ReachingDefinitionsAnalysis(
 
     def _post_analysis(self):
         pass
+
+    def callsites_to(self, target: Union[int, str, Function]) -> Iterable[FunctionCallRelationships]:
+        if isinstance(target, (str, int)):
+            try:
+                func_addr = self.project.kb.functions[target].addr
+            except KeyError:
+                return
+        elif isinstance(target, Function):
+            func_addr = target.addr
+        else:
+            raise TypeError(type(target))
+
+        for info in self.function_calls.values():
+            if info.target == func_addr:
+                yield info
