@@ -20,13 +20,13 @@ class TypeTranslator:
     """
     Translate type variables to SimType equivalence.
     """
-    def __init__(self, arch=None):
 
+    def __init__(self, arch=None):
         self.arch = arch
 
-        self.translated: Dict[TypeConstant,SimType] = { }
-        self.translated_simtypes: Dict[SimType,TypeConstant] = { }
-        self.structs = { }
+        self.translated: Dict[TypeConstant, SimType] = {}
+        self.translated_simtypes: Dict[SimType, TypeConstant] = {}
+        self.structs = {}
         self._struct_ctr = count()
 
         # will be updated every time .translate() is called
@@ -37,7 +37,6 @@ class TypeTranslator:
     #
 
     def struct_name(self):
-
         return "struct_%d" % next(self._struct_ctr)
 
     #
@@ -45,12 +44,10 @@ class TypeTranslator:
     #
 
     def tc2simtype(self, tc):
-
         self._has_nonexistent_ref = False
         return self._tc2simtype(tc), self._has_nonexistent_ref
 
     def _tc2simtype(self, tc):
-
         if tc is None:
             return sim_type.SimTypeBottom().with_arch(self.arch)
 
@@ -79,7 +76,6 @@ class TypeTranslator:
     #
 
     def _translate_Pointer64(self, tc):
-
         if isinstance(tc.basetype, typeconsts.BottomType):
             # void *
             internal = sim_type.SimTypeBottom(label="void").with_arch(self.arch)
@@ -88,7 +84,6 @@ class TypeTranslator:
         return sim_type.SimTypePointer(internal).with_arch(self.arch)
 
     def _translate_Pointer32(self, tc):
-
         if isinstance(tc.basetype, typeconsts.BottomType):
             # void *
             internal = sim_type.SimTypeBottom(label="void").with_arch(self.arch)
@@ -96,16 +91,20 @@ class TypeTranslator:
             internal = self._tc2simtype(tc.basetype)
         return sim_type.SimTypePointer(internal).with_arch(self.arch)
 
-    def _translate_Array(self, tc: typeconsts.Array):
+    def _translate_Array(self, tc: typeconsts.Array) -> sim_type.SimTypeArray:
         elem_type = self._tc2simtype(tc.element)
         return sim_type.SimTypeArray(elem_type, length=tc.count).with_arch(self.arch)
 
-    def _translate_Struct(self, tc):
-
+    def _translate_Struct(self, tc: typeconsts.Struct):
         if tc in self.structs:
             return self.structs[tc]
 
-        s = sim_type.SimStruct({}, name=self.struct_name()).with_arch(self.arch)
+        if tc.name:
+            name = tc.name
+        else:
+            name = self.struct_name()
+
+        s = sim_type.SimStruct({}, name=name).with_arch(self.arch)
         self.structs[tc] = s
 
         next_offset = 0
@@ -123,7 +122,11 @@ class TypeTranslator:
                 # for now, we replace it with an unsigned char
                 translated_type = sim_type.SimTypeChar(signed=False).with_arch(self.arch)
 
-            s.fields["field_%x" % offset] = translated_type
+            if tc.field_names and offset in tc.field_names:
+                field_name = tc.field_names[offset]
+            else:
+                field_name = f"field_{offset:x}"
+            s.fields[field_name] = translated_type
 
             if isinstance(translated_type, SimTypeTempRef):
                 next_offset = self.arch.bytes + offset
@@ -148,7 +151,6 @@ class TypeTranslator:
         return sim_type.SimTypeNum(128, signed=False).with_arch(self.arch)
 
     def _translate_TypeVariableReference(self, tc):
-
         if tc.typevar in self.translated:
             return self.translated[tc.typevar]
 
@@ -171,7 +173,7 @@ class TypeTranslator:
             self.backpatch(st.pts_to, translated)
 
         elif isinstance(st, sim_type.SimStruct):
-            fields_patch = { }
+            fields_patch = {}
             for offset, fld in st.fields.items():
                 if isinstance(fld, SimTypeTempRef) and fld.typevar in translated:
                     fields_patch[offset] = translated[fld.typevar]
@@ -194,7 +196,7 @@ class TypeTranslator:
         return typeconsts.Int8()
 
     def _translate_SimStruct(self, st: sim_type.SimStruct) -> typeconsts.Struct:
-        fields = { }
+        fields = {}
         offsets = st.offsets
         for name, ty in st.fields.items():
             offset = offsets[name]
@@ -207,8 +209,9 @@ class TypeTranslator:
         array_tc = typeconsts.Array(elem_type, count=st.length)
         return array_tc
 
-    def _translate_SimTypePointer(self,
-                                  st: sim_type.SimTypePointer) -> Union[typeconsts.Pointer32, typeconsts.Pointer64]:
+    def _translate_SimTypePointer(
+        self, st: sim_type.SimTypePointer
+    ) -> Union[typeconsts.Pointer32, typeconsts.Pointer64]:
         base = self._simtype2tc(st.pts_to)
         if self.arch.bits == 32:
             return typeconsts.Pointer32(base)
