@@ -2,11 +2,14 @@
 from typing import Optional, TYPE_CHECKING
 import logging
 
-import claripy
 import ailment
+import claripy
+from unique_log_filter import UniqueLogFilter
 
+from angr.procedures import SIM_LIBRARIES, SIM_TYPE_COLLECTIONS
+from angr.utils.constants import MAX_POINTSTO_BITS
 from ...calling_conventions import SimRegArg
-from ...sim_type import SimTypeFunction, SimTypeBottom
+from ...sim_type import SimTypeFunction, dereference_simtype
 from ...engines.light import SimEngineLightAILMixin
 from ..typehoon import typeconsts, typevars
 from ..typehoon.lifter import TypeLifter
@@ -17,13 +20,19 @@ if TYPE_CHECKING:
 
 
 l = logging.getLogger(name=__name__)
+l.addFilter(UniqueLogFilter())
 
 
 class SimEngineVRAIL(
     SimEngineLightAILMixin,
     SimEngineVRBase,
 ):
+    """
+    The engine for variable recovery on AIL.
+    """
+
     state: "VariableRecoveryFastState"
+    block: ailment.Block
 
     def __init__(self, *args, call_info=None, **kwargs):
         super().__init__(*args, **kwargs)
@@ -41,6 +50,12 @@ class SimEngineVRAIL(
             data = self._expr(stmt.src)
             size = stmt.src.bits // 8
 
+            if hasattr(stmt.dst, "write_size") and stmt.dst.write_size > size:
+                # zero-fill this register
+                self._assign_to_register(
+                    offset, RichR(self.state.top(stmt.dst.write_size * 8)), stmt.dst.write_size, create_variable=False
+                )
+
             self._assign_to_register(offset, data, size, src=stmt.src, dst=stmt.dst)
 
         elif dst_type is ailment.Expr.Tmp:
@@ -80,6 +95,7 @@ class SimEngineVRAIL(
         ret_reg_offset = None
         ret_expr_bits = self.state.arch.bits
         ret_val = None  # stores the value that this method should return to its caller when this is a call expression.
+        create_variable = True
         if not is_expr:
             # this is a call statement. we need to update the return value register later
             ret_expr: Optional[ailment.Expr.Register] = stmt.ret_expr
@@ -87,13 +103,11 @@ class SimEngineVRAIL(
                 ret_reg_offset = ret_expr.reg_offset
                 ret_expr_bits = ret_expr.bits
             else:
+                # the return expression is not used, so we treat this call as not returning anything
                 if stmt.calling_convention is not None:
+                    # we only set the ret_expr if prototype must be guessed. otherwise ret_expr should just be None
                     if stmt.prototype is None:
                         ret_expr: SimRegArg = stmt.calling_convention.RETURN_VAL
-                    elif stmt.prototype.returnty is None or type(stmt.prototype.returnty) is SimTypeBottom:
-                        ret_expr = None
-                    else:
-                        ret_expr: SimRegArg = stmt.calling_convention.return_val(stmt.prototype.returnty)
                 else:
                     l.debug(
                         "Unknown calling convention for function %s. Fall back to default calling convention.", target
@@ -102,20 +116,31 @@ class SimEngineVRAIL(
 
                 if ret_expr is not None:
                     ret_reg_offset = self.project.arch.registers[ret_expr.reg_name][0]
+                create_variable = False
         else:
             # this is a call expression. we just return the value at the end of this method
             if stmt.ret_expr is not None:
                 ret_expr_bits = stmt.ret_expr.bits
 
+        if isinstance(target, ailment.Expr.Expression) and not isinstance(target, ailment.Expr.Const):
+            # this is a dynamically calculated call target
+            target_expr = self._expr(target)
+            funcaddr_typevar = target_expr.typevar
+            load_typevar = self._create_access_typevar(target_expr.typevar, False, self.arch.bytes, 0)
+            self.state.add_type_constraint(typevars.Subtype(funcaddr_typevar, load_typevar))
+
         # discover the prototype
         prototype: Optional[SimTypeFunction] = None
+        prototype_libname: Optional[str] = None
         if stmt.prototype is not None:
             prototype = stmt.prototype
-        elif isinstance(stmt.target, ailment.Expr.Const):
+        if isinstance(stmt.target, ailment.Expr.Const):
             func_addr = stmt.target.value
-            if func_addr in self.kb.functions:
+            if isinstance(func_addr, self.kb.functions.address_types) and func_addr in self.kb.functions:
                 func = self.kb.functions[func_addr]
-                prototype = func.prototype
+                if prototype is None:
+                    prototype = func.prototype
+                prototype_libname = func.prototype_libname
 
         # dump the type of the return value
         if prototype is not None:
@@ -141,14 +166,24 @@ class SimEngineVRAIL(
                 self._assign_to_register(
                     ret_reg_offset,
                     RichR(self.state.top(expr_bits), typevar=ret_ty),
-                    self.state.arch.bytes,
+                    expr_bits // self.arch.byte_width,
                     dst=ret_expr,
+                    create_variable=create_variable,
                 )
 
         if prototype is not None and args:
             # add type constraints
+
+            type_collections = []
+            if prototype_libname is not None:
+                prototype_lib = SIM_LIBRARIES[prototype_libname]
+                if prototype_lib.type_collection_names:
+                    for typelib_name in prototype_lib.type_collection_names:
+                        type_collections.append(SIM_TYPE_COLLECTIONS[typelib_name])
+
             for arg, arg_type in zip(args, prototype.args):
                 if arg.typevar is not None:
+                    arg_type = dereference_simtype(arg_type, type_collections).with_arch(arg_type._arch)
                     arg_ty = TypeLifter(self.arch.bits).lift(arg_type)
                     type_constraint = typevars.Subtype(arg.typevar, arg_ty)
                     self.state.add_type_constraint(type_constraint)
@@ -197,7 +232,7 @@ class SimEngineVRAIL(
         r = self._load(addr_r, size, expr=expr)
         return r
 
-    def _ail_handle_Const(self, expr):
+    def _ail_handle_Const(self, expr: ailment.Expr.Const):
         if isinstance(expr.value, float):
             v = claripy.FPV(expr.value, claripy.FSORT_DOUBLE if expr.bits == 64 else claripy.FSORT_FLOAT)
             ty = typeconsts.float_type(expr.bits)
@@ -205,11 +240,20 @@ class SimEngineVRAIL(
             if self.project.loader.find_segment_containing(expr.value) is not None:
                 r = self._load_from_global(expr.value, 1, expr=expr)
                 ty = r.typevar
+            elif expr.value == 0 and expr.bits == self.arch.bits:
+                # this can be viewed as a NULL
+                ty = (
+                    typeconsts.Pointer64(typeconsts.TopType())
+                    if self.arch.bits == 64
+                    else typeconsts.Pointer32(typeconsts.TopType())
+                )
             else:
                 ty = typeconsts.int_type(expr.size * self.state.arch.byte_width)
-            v = claripy.BVV(expr.value, expr.size * self.state.arch.byte_width)
+            v = claripy.BVV(expr.value, expr.bits)
         r = RichR(v, typevar=ty)
-        self._reference(r, self._codeloc())
+        codeloc = self._codeloc()
+        self._ensure_variable_existence(r, codeloc)
+        self._reference(r, codeloc)
         return r
 
     def _ail_handle_BinaryOp(self, expr):
@@ -226,7 +270,9 @@ class SimEngineVRAIL(
         r = self._expr(expr.operand)
         typevar = None
         if r.typevar is not None:
-            if isinstance(r.typevar, typevars.DerivedTypeVariable) and isinstance(r.typevar.label, typevars.ConvertTo):
+            if isinstance(r.typevar, typevars.DerivedTypeVariable) and isinstance(
+                r.typevar.one_label, typevars.ConvertTo
+            ):
                 # there is already a conversion - overwrite it
                 if not isinstance(r.typevar.type_var, typeconsts.TypeConstant):
                     typevar = typevars.DerivedTypeVariable(r.typevar.type_var, typevars.ConvertTo(expr.to_bits))
@@ -241,7 +287,7 @@ class SimEngineVRAIL(
         typevar = None
         if r.typevar is not None:
             if isinstance(r.typevar, typevars.DerivedTypeVariable) and isinstance(
-                r.typevar.label, typevars.ReinterpretAs
+                r.typevar.one_label, typevars.ReinterpretAs
             ):
                 # there is already a reinterpretas - overwrite it
                 typevar = typevars.DerivedTypeVariable(
@@ -253,17 +299,27 @@ class SimEngineVRAIL(
         return RichR(self.state.top(expr.to_bits), typevar=typevar)
 
     def _ail_handle_StackBaseOffset(self, expr: ailment.Expr.StackBaseOffset):
-        typevar = self.state.stack_offset_typevars.get(expr.offset, None)
+        ref_typevar = self.state.stack_offset_typevars.get(expr.offset, None)
 
-        if typevar is None:
+        if ref_typevar is None:
             # allocate a new type variable
-            typevar = typevars.TypeVariable()
-            self.state.stack_offset_typevars[expr.offset] = typevar
+            ref_typevar = typevars.TypeVariable()
+            self.state.stack_offset_typevars[expr.offset] = ref_typevar
 
         value_v = self.state.stack_address(expr.offset)
-        richr = RichR(value_v, typevar=typevar)
+        richr = RichR(value_v, typevar=ref_typevar)
+        codeloc = self._codeloc()
+        var_and_offsets = self._ensure_variable_existence(richr, codeloc, src_expr=expr)
         if self._reference_spoffset:
-            self._reference(richr, self._codeloc(), src=expr)
+            self._reference(richr, codeloc, src=expr)
+        for var, off_in_var in var_and_offsets:
+            if self.state.typevars.has_type_variable_for(var, codeloc):
+                var_typevar = self.state.typevars.get_type_variable(var, codeloc)
+                load_typevar = self._create_access_typevar(
+                    ref_typevar, False, MAX_POINTSTO_BITS // 8, 0 if off_in_var is None else off_in_var
+                )
+                type_constraint = typevars.Subtype(var_typevar, load_typevar)
+                self.state.add_type_constraint(type_constraint)
 
         return richr
 
@@ -302,7 +358,7 @@ class SimEngineVRAIL(
 
         if r1.data.concrete:
             # addition with constants. create a derived type variable
-            typevar = typevars.DerivedTypeVariable(r0_typevar, typevars.AddN(r1.data._model_concrete.value))
+            typevar = typevars.DerivedTypeVariable(r0_typevar, typevars.AddN(r1.data.concrete_value))
         elif r1.typevar is not None:
             typevar = typevars.TypeVariable()
             type_constraints.add(typevars.Add(r0_typevar, r1.typevar, typevar))
@@ -327,7 +383,7 @@ class SimEngineVRAIL(
 
         type_constraints = set()
         if r0.typevar is not None and r1.data.concrete:
-            typevar = typevars.DerivedTypeVariable(r0.typevar, typevars.SubN(r1.data._model_concrete.value))
+            typevar = typevars.DerivedTypeVariable(r0.typevar, typevars.SubN(r1.data.concrete_value))
         else:
             typevar = typevars.TypeVariable()
             if r0.typevar is not None and r1.typevar is not None:
@@ -396,15 +452,15 @@ class SimEngineVRAIL(
         to_size = r1.bits
 
         if expr.floating_point:
-            remainder = self.state.top(to_size)
+            quotient = self.state.top(to_size)
         else:
             if expr.signed:
-                remainder = r0.data.SMod(claripy.SignExt(from_size - to_size, r1.data))
+                quotient = claripy.SDiv(r0.data, claripy.SignExt(from_size - to_size, r1.data))
             else:
-                remainder = r0.data % claripy.ZeroExt(from_size - to_size, r1.data)
+                quotient = r0.data / claripy.ZeroExt(from_size - to_size, r1.data)
 
         return RichR(
-            remainder,
+            quotient,
             # | typevar=r0.typevar,  # FIXME: Handle typevars for Div
         )
 
@@ -413,20 +469,20 @@ class SimEngineVRAIL(
 
         r0 = self._expr(arg0)
         r1 = self._expr(arg1)
-        from_size = r1.bits
-        to_size = r0.bits
+        from_size = expr.from_bits
+        to_size = expr.to_bits
 
         if expr.signed:
-            quotient = r0.data.SDiv(claripy.SignExt(to_size - from_size, r1.data))
-            remainder = r0.data.SMod(claripy.SignExt(to_size - from_size, r1.data))
+            quotient = r0.data.SDiv(claripy.SignExt(from_size - to_size, r1.data))
+            remainder = r0.data.SMod(claripy.SignExt(from_size - to_size, r1.data))
             quotient_size = to_size
             remainder_size = to_size
             r = claripy.Concat(
                 claripy.Extract(remainder_size - 1, 0, remainder), claripy.Extract(quotient_size - 1, 0, quotient)
             )
         else:
-            quotient = r0.data // claripy.ZeroExt(to_size - from_size, r1.data)
-            remainder = r0.data % claripy.ZeroExt(to_size - from_size, r1.data)
+            quotient = r0.data // claripy.ZeroExt(from_size - to_size, r1.data)
+            remainder = r0.data % claripy.ZeroExt(from_size - to_size, r1.data)
             quotient_size = to_size
             remainder_size = to_size
             r = claripy.Concat(
@@ -438,6 +494,27 @@ class SimEngineVRAIL(
             # | typevar=r0.typevar,  # FIXME: Handle typevars for DivMod
         )
 
+    def _ail_handle_Mod(self, expr):
+        arg0, arg1 = expr.operands
+
+        r0 = self._expr(arg0)
+        r1 = self._expr(arg1)
+        from_size = expr.bits
+        to_size = r1.bits
+
+        if expr.floating_point:
+            remainder = self.state.top(to_size)
+        else:
+            if expr.signed:
+                remainder = r0.data.SMod(claripy.SignExt(from_size - to_size, r1.data))
+            else:
+                remainder = r0.data % claripy.ZeroExt(from_size - to_size, r1.data)
+
+        return RichR(
+            remainder,
+            # | typevar=r0.typevar,  # FIXME: Handle typevars for Mod
+        )
+
     def _ail_handle_Xor(self, expr):
         arg0, arg1 = expr.operands
 
@@ -470,7 +547,7 @@ class SimEngineVRAIL(
                 typevar=r0.typevar,
             )
 
-        shiftamount = r1.data._model_concrete.value
+        shiftamount = r1.data.concrete_value
 
         return RichR(r0.data << shiftamount, typevar=typeconsts.int_type(result_size), type_constraints=None)
 
@@ -489,7 +566,7 @@ class SimEngineVRAIL(
                 typevar=r0.typevar,
             )
 
-        shiftamount = r1.data._model_concrete.value
+        shiftamount = r1.data.concrete_value
 
         return RichR(
             claripy.LShR(r0.data, shiftamount), typevar=typeconsts.int_type(result_size), type_constraints=None
@@ -510,7 +587,7 @@ class SimEngineVRAIL(
                 typevar=r0.typevar,
             )
 
-        shiftamount = r1.data._model_concrete.value
+        shiftamount = r1.data.concrete_value
 
         return RichR(r0.data << shiftamount, typevar=typeconsts.int_type(result_size), type_constraints=None)
 
@@ -529,7 +606,7 @@ class SimEngineVRAIL(
                 typevar=r0.typevar,
             )
 
-        shiftamount = r1.data._model_concrete.value
+        shiftamount = r1.data.concrete_value
 
         return RichR(r0.data >> shiftamount, typevar=typeconsts.int_type(result_size), type_constraints=None)
 
@@ -567,6 +644,50 @@ class SimEngineVRAIL(
         r = self.state.top(expr.bits)
         return RichR(r, typevar=r0.typevar)
 
+    def _ail_handle_LogicalAnd(self, expr):
+        arg0, arg1 = expr.operands
+
+        r0 = self._expr(arg0)
+        _ = self._expr(arg1)
+        r = self.state.top(expr.bits)
+        return RichR(r, typevar=r0.typevar)
+
+    def _ail_handle_LogicalOr(self, expr):
+        arg0, arg1 = expr.operands
+
+        r0 = self._expr(arg0)
+        _ = self._expr(arg1)
+        r = self.state.top(expr.bits)
+        return RichR(r, typevar=r0.typevar)
+
+    def _ail_handle_LogicalXor(self, expr):
+        arg0, arg1 = expr.operands
+
+        r0 = self._expr(arg0)
+        _ = self._expr(arg1)
+        r = self.state.top(expr.bits)
+        return RichR(r, typevar=r0.typevar)
+
+    def _ail_handle_Rol(self, expr):
+        arg0, arg1 = expr.operands
+
+        r0 = self._expr(arg0)
+        _ = self._expr(arg1)
+        result_size = arg0.bits
+
+        r = self.state.top(result_size)
+        return RichR(r, typevar=r0.typevar)
+
+    def _ail_handle_Ror(self, expr):
+        arg0, arg1 = expr.operands
+
+        r0 = self._expr(arg0)
+        _ = self._expr(arg1)
+        result_size = arg0.bits
+
+        r = self.state.top(result_size)
+        return RichR(r, typevar=r0.typevar)
+
     def _ail_handle_Concat(self, expr):
         arg0, arg1 = expr.operands
 
@@ -607,3 +728,19 @@ class SimEngineVRAIL(
 
         r = self.state.top(result_size)
         return RichR(r, typevar=expr.typevar)
+
+    def _ail_handle_BitwiseNeg(self, expr):
+        arg = expr.operands[0]
+        expr = self._expr(arg)
+
+        result_size = arg.bits
+
+        if expr.data.concrete:
+            return RichR(
+                ~expr.data,
+                typevar=typeconsts.int_type(result_size),
+                type_constraints=None,
+            )
+
+        r = self.state.top(result_size)
+        return RichR(r, typevar=expr.typevar)
