@@ -118,6 +118,26 @@ class VariableRecoveryBase(Analysis):
             for d in domfront:
                 self._dominance_frontiers[d.addr].add(b0.addr)
 
+    def _post_analysis(self):
+        # remove temporary variables (stack variables created by _ensure_variable_existence() that are 1-byte long,
+        # never accessed, and overlap with other stack variables at the same offset)
+        varman = self.variable_manager[self.function.addr]
+        stack_vars = varman.get_variables("stack")
+        stack_vars_by_offset = defaultdict(list)
+        for sv in stack_vars:
+            stack_vars_by_offset[sv.offset].append(sv)
+        for offset, var_list in stack_vars_by_offset.items():
+            if len(var_list) < 2:
+                continue
+            single_byte_vars = [v for v in var_list if v.size == 1]
+            if len(single_byte_vars) != 1:
+                continue
+            single_byte_var = single_byte_vars[0]
+
+            if not varman.get_variable_accesses(single_byte_var):
+                # remove this variable
+                varman._variables.discard(single_byte_var)
+
 
 class VariableRecoveryStateBase:
     """
@@ -137,6 +157,7 @@ class VariableRecoveryStateBase:
         global_region=None,
         typevars=None,
         type_constraints=None,
+        func_typevar=None,
         delayed_type_constraints=None,
         stack_offset_typevars=None,
         project=None,
@@ -154,6 +175,7 @@ class VariableRecoveryStateBase:
             self.stack_region: MultiValuedMemory = MultiValuedMemory(
                 memory_id="mem",
                 top_func=self.top,
+                is_top_func=self.is_top,
                 phi_maker=self._make_phi_variable,
                 skip_missing_values_during_merging=True,
                 page_kwargs={"mo_cmp": self._mo_cmp},
@@ -167,6 +189,7 @@ class VariableRecoveryStateBase:
             self.register_region: MultiValuedMemory = MultiValuedMemory(
                 memory_id="reg",
                 top_func=self.top,
+                is_top_func=self.is_top,
                 phi_maker=self._make_phi_variable,
                 skip_missing_values_during_merging=True,
                 page_kwargs={"mo_cmp": self._mo_cmp},
@@ -180,6 +203,7 @@ class VariableRecoveryStateBase:
             self.global_region: MultiValuedMemory = MultiValuedMemory(
                 memory_id="mem",
                 top_func=self.top,
+                is_top_func=self.is_top,
                 phi_maker=self._make_phi_variable,
                 skip_missing_values_during_merging=True,
                 page_kwargs={"mo_cmp": self._mo_cmp},
@@ -191,9 +215,10 @@ class VariableRecoveryStateBase:
         self.phi_variables: Dict[SimVariable, SimVariable] = {}
 
         self.typevars = TypeVariables() if typevars is None else typevars
-        self.type_constraints = set() if type_constraints is None else type_constraints
+        self.type_constraints = defaultdict(set) if type_constraints is None else type_constraints
+        self.func_typevar = func_typevar
         self.delayed_type_constraints = (
-            DefaultChainMapCOW(set, collapse_threshold=25)
+            DefaultChainMapCOW(default_factory=set, collapse_threshold=25)
             if delayed_type_constraints is None
             else delayed_type_constraints
         )
@@ -243,7 +268,7 @@ class VariableRecoveryStateBase:
 
     def is_global_variable_address(self, addr: claripy.ast.Base) -> bool:
         if addr.op == "BVV":
-            addr_v = addr._model_concrete.value
+            addr_v = addr.concrete_value
             # make sure it is within a mapped region
             obj = self.project.loader.find_object_containing(addr_v)
             if obj is not None:
@@ -251,32 +276,40 @@ class VariableRecoveryStateBase:
         return False
 
     @staticmethod
-    def extract_stack_offset_from_addr(addr: claripy.ast.Base) -> claripy.ast.Base:
+    def extract_stack_offset_from_addr(addr: claripy.ast.Base) -> Optional[claripy.ast.Base]:
         r = None
         if addr.op == "BVS":
-            r = claripy.BVV(0, addr.size())
+            if addr.args[0] == "stack_base":
+                return claripy.BVV(0, addr.size())
+            return None
         elif addr.op == "BVV":
             r = addr
         elif addr.op == "__add__":
-            r = sum(VariableRecoveryStateBase.extract_stack_offset_from_addr(arg) for arg in addr.args)
+            arg_offsets = []
+            for arg in addr.args:
+                arg_offset = VariableRecoveryStateBase.extract_stack_offset_from_addr(arg)
+                if arg_offset is None:
+                    return None
+                arg_offsets.append(arg_offset)
+            r = sum(arg_offsets)
         elif addr.op == "__sub__":
             r1 = VariableRecoveryStateBase.extract_stack_offset_from_addr(addr.args[0])
             r2 = VariableRecoveryStateBase.extract_stack_offset_from_addr(addr.args[1])
+            if r1 is None or r2 is None:
+                return None
             r = r1 - r2
-        else:
-            # NOTE: The original code here didn't support mul or
-            # anything like that, so let's specify it as 0
-            r = claripy.BVV(0, addr.size())
         return r
 
     def get_stack_offset(self, addr: claripy.ast.Base) -> Optional[int]:
         if "stack_base" in addr.variables:
             r = VariableRecoveryStateBase.extract_stack_offset_from_addr(addr)
+            if r is None:
+                return None
 
             # extract_stack_offset_from_addr should ensure that r is a BVV
             assert r.concrete
 
-            val = r._model_concrete.value
+            val = r.concrete_value
             # convert it to a signed integer
             if val >= 2 ** (self.arch.bits - 1):
                 return val - 2**self.arch.bits
@@ -334,7 +367,18 @@ class VariableRecoveryStateBase:
         :return:
         """
 
-        self.type_constraints.add(constraint)
+        self.type_constraints[self.func_typevar].add(constraint)
+
+    def add_type_constraint_for_function(self, func_typevar, constraint):
+        """
+        Add a new type constraint for a specified function.
+
+        :param func_typevar:
+        :param constraint:
+        :return:
+        """
+
+        self.type_constraints[func_typevar].add(constraint)
 
     def downsize(self) -> None:
         """
@@ -342,7 +386,7 @@ class VariableRecoveryStateBase:
 
         :return:    None
         """
-        self.type_constraints = set()
+        self.type_constraints = defaultdict(set)
 
     @staticmethod
     def downsize_region(region: MultiValuedMemory) -> MultiValuedMemory:
