@@ -1,5 +1,5 @@
 # pylint:disable=wrong-import-position,wrong-import-order
-from typing import Optional, List, Tuple
+from typing import Optional, List, Tuple, Union, DefaultDict, Set, Dict, TYPE_CHECKING
 import logging
 from collections import defaultdict
 
@@ -16,12 +16,15 @@ from ...errors import AngrVariableRecoveryError, SimEngineError
 from ...knowledge_plugins import Function
 from ...sim_variable import SimStackVariable, SimRegisterVariable, SimVariable, SimMemoryVariable
 from ...engines.vex.claripy.irop import vexop_to_simop
-from ..forward_analysis import ForwardAnalysis, FunctionGraphVisitor
-from ..typehoon.typevars import Equivalence, TypeVariable
+from angr.analyses import ForwardAnalysis, visitors
+from ..typehoon.typevars import Equivalence, TypeVariable, TypeVariables
 from .variable_recovery_base import VariableRecoveryBase, VariableRecoveryStateBase
 from .engine_vex import SimEngineVRVEX
 from .engine_ail import SimEngineVRAIL
 
+if TYPE_CHECKING:
+    from angr.analyses.typehoon.typevars import TypeConstraint
+
 l = logging.getLogger(name=__name__)
 
 
@@ -44,6 +47,7 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
         global_region=None,
         typevars=None,
         type_constraints=None,
+        func_typevar=None,
         delayed_type_constraints=None,
         stack_offset_typevars=None,
         project=None,
@@ -59,6 +63,7 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
             global_region=global_region,
             typevars=typevars,
             type_constraints=type_constraints,
+            func_typevar=func_typevar,
             delayed_type_constraints=delayed_type_constraints,
             stack_offset_typevars=stack_offset_typevars,
             project=project,
@@ -86,9 +91,10 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
             stack_region=self.stack_region.copy(),
             register_region=self.register_region.copy(),
             global_region=self.global_region.copy(),
-            typevars=self.typevars.copy(),
-            type_constraints=self.type_constraints.copy(),
-            delayed_type_constraints=self.delayed_type_constraints.copy(),
+            typevars=self.typevars,
+            type_constraints=self.type_constraints,
+            func_typevar=self.func_typevar,
+            delayed_type_constraints=self.delayed_type_constraints,
             stack_offset_typevars=dict(self.stack_offset_typevars),
             project=self.project,
             ret_val_size=self.ret_val_size,
@@ -125,26 +131,17 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
         merged_global_region.set_state(self)
         merge_occurred |= merged_global_region.merge([other.global_region for other in others], None)
 
-        merged_typevars = self.typevars
-        merged_typeconstraints = self.type_constraints.copy()
-        delayed_typeconstraints = self.delayed_type_constraints.copy().clean()
-        for other in others:
-            merged_typevars = merged_typevars.merge(other.typevars)
-            merged_typeconstraints |= other.type_constraints
-            for v, cons in other.delayed_type_constraints.items():
-                delayed_typeconstraints[v] |= cons
-
-        merge_occurred |= self.typevars != merged_typevars
-        merge_occurred |= self.type_constraints != merged_typeconstraints
-        merge_occurred |= self.delayed_type_constraints != delayed_typeconstraints
+        typevars = self.typevars
+        type_constraints = self.type_constraints
+        delayed_typeconstraints = self.delayed_type_constraints
 
         # add subtype constraints for all replacements
         for v0, v1 in self.phi_variables.items():
             # v0 will be replaced by v1
-            if not merged_typevars.has_type_variable_for(v1, None):
-                merged_typevars.add_type_variable(v1, None, TypeVariable())
-            if not merged_typevars.has_type_variable_for(v0, None):
-                merged_typevars.add_type_variable(v0, None, TypeVariable())
+            if not typevars.has_type_variable_for(v1, None):
+                typevars.add_type_variable(v1, None, TypeVariable())
+            if not typevars.has_type_variable_for(v0, None):
+                typevars.add_type_variable(v0, None, TypeVariable())
             # Assuming v2 = phi(v0, v1), then we know that v0_typevar == v1_typevar == v2_typevar
             # However, it's possible that neither v0 nor v1 will ever be used in future blocks, which not only makes
             # this phi function useless, but also leads to the incorrect assumption that v1_typevar == v2_typevar.
@@ -152,9 +149,7 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
             # when v1 (the new variable that will end up in the state) is ever used in the future.
 
             # create an equivalence relationship
-            equivalence = Equivalence(
-                merged_typevars.get_type_variable(v1, None), merged_typevars.get_type_variable(v0, None)
-            )
+            equivalence = Equivalence(typevars.get_type_variable(v1, None), typevars.get_type_variable(v0, None))
             delayed_typeconstraints[v1].add(equivalence)
 
         stack_offset_typevars = {}
@@ -173,7 +168,7 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
             else:
                 typevar = TypeVariable()
                 for orig_typevar in all_typevars:
-                    merged_typeconstraints.add(Equivalence(orig_typevar, typevar))
+                    type_constraints[self.func_typevar].add(Equivalence(orig_typevar, typevar))
             stack_offset_typevars[offset] = typevar
 
         ret_val_size = self.ret_val_size
@@ -195,8 +190,9 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
             stack_region=merged_stack_region,
             register_region=merged_register_region,
             global_region=merged_global_region,
-            typevars=merged_typevars,
-            type_constraints=merged_typeconstraints,
+            typevars=typevars,
+            type_constraints=type_constraints,
+            func_typevar=self.func_typevar,
             delayed_type_constraints=delayed_typeconstraints,
             stack_offset_typevars=stack_offset_typevars,
             project=self.project,
@@ -205,6 +201,9 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
 
         return state, merge_occurred
 
+    def downsize(self) -> None:
+        pass
+
     #
     # Util methods
     #
@@ -234,14 +233,17 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
 
     def __init__(
         self,
-        func: Function,
+        func: Union[Function, str, int],
         func_graph: Optional[networkx.DiGraph] = None,
         max_iterations: int = 2,
         low_priority=False,
         track_sp=True,
         func_args: Optional[List[SimVariable]] = None,
         store_live_variables=False,
+        unify_variables=True,
     ):
+        if not isinstance(func, Function):
+            func = self.kb.functions[func]
         func_graph_with_calls = func_graph or func.transition_graph
         call_info = defaultdict(list)
         for node_from, node_to, data in func_graph_with_calls.edges(data=True):
@@ -251,7 +253,7 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
                 except KeyError:
                     pass
 
-        function_graph_visitor = FunctionGraphVisitor(func, graph=func_graph)
+        function_graph_visitor = visitors.FunctionGraphVisitor(func, graph=func_graph)
 
         # Make sure the function is not empty
         if not func.block_addrs_set or func.startpoint is None:
@@ -266,6 +268,7 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
         self._job_ctr = 0
         self._track_sp = track_sp and self.project.arch.sp_offset is not None
         self._func_args = func_args
+        self._unify_variables = unify_variables
 
         self._ail_engine = SimEngineVRAIL(self.project, self.kb, call_info=call_info)
         self._vex_engine = SimEngineVRVEX(self.project, self.kb, call_info=call_info)
@@ -273,8 +276,11 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
         self._node_iterations = defaultdict(int)
 
         self._node_to_cc = {}
-        self.var_to_typevars = defaultdict(set)
-        self.type_constraints = None
+        self.var_to_typevars: DefaultDict[SimVariable, Set[TypeVariable]] = defaultdict(set)
+        self.typevars = None
+        self.type_constraints: Optional[Dict["TypeVariable", Set["TypeConstraint"]]] = None
+        self.func_typevar = TypeVariable(name=func.name)
+        self.delayed_type_constraints = None
         self.ret_val_size = None
 
         self._analyze()
@@ -289,7 +295,9 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
     #
 
     def _pre_analysis(self):
-        self.type_constraints = set()
+        self.typevars = TypeVariables()
+        self.type_constraints = defaultdict(set)
+        self.delayed_type_constraints = defaultdict(set)
 
         self.initialize_dominance_frontiers()
 
@@ -317,6 +325,10 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
             self.project.arch,
             self.function,
             project=self.project,
+            typevars=self.typevars,
+            type_constraints=self.type_constraints,
+            func_typevar=self.func_typevar,
+            delayed_type_constraints=self.delayed_type_constraints,
         )
         initial_sp = state.stack_address(self.project.arch.bytes if self.project.arch.call_pushes_ret else 0)
         if self.project.arch.sp_offset is not None:
@@ -330,15 +342,25 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
         # put a return address on the stack if necessary
         if self.project.arch.call_pushes_ret:
             ret_addr_offset = self.project.arch.bytes
-            ret_addr_var = SimStackVariable(
-                ret_addr_offset,
-                self.project.arch.bytes,
-                base="bp",
-                name="ret_addr",
-                region=self.function.addr,
-                category="return_address",
-                ident=internal_manager.next_variable_ident("stack"),
+            # find existing variable
+            ret_addr_var = next(
+                iter(
+                    v
+                    for v in internal_manager.find_variables_by_stack_offset(ret_addr_offset)
+                    if v.category == "return_address"
+                ),
+                None,
             )
+            if ret_addr_var is None:
+                ret_addr_var = SimStackVariable(
+                    ret_addr_offset,
+                    self.project.arch.bytes,
+                    base="bp",
+                    name="ret_addr",
+                    region=self.function.addr,
+                    category="return_address",
+                    ident=internal_manager.next_variable_ident("stack"),
+                )
             ret_addr = claripy.BVS("ret_addr", self.project.arch.bits)
             ret_addr = state.annotate_with_variables(ret_addr, [(0, ret_addr_var)])
             state.stack_region.store(
@@ -392,9 +414,11 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
         if type(node) is ailment.Block:
             # AIL mode
             block = node
+            block_key = node.addr, node.idx
         else:
             # VEX mode, get the block again
             block = self.project.factory.block(node.addr, node.size, opt_level=1, cross_insn_opt=False)
+            block_key = node.addr
 
         # if node.addr in self._instates:
         #     prev_state: VariableRecoveryFastState = self._instates[node.addr]
@@ -407,25 +431,24 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
 
         state = state.copy()
         state.block_addr = node.addr
+        if isinstance(node, ailment.Block):
+            state.block_idx = node.idx
         # self._instates[node.addr] = state
 
-        if self._node_iterations[node.addr] >= self._max_iterations:
+        if self._node_iterations[block_key] >= self._max_iterations:
             l.debug("Skip node %#x as we have iterated %d times on it.", node.addr, self._node_iterations[node.addr])
             return False, state
 
         self._process_block(state, block)
 
-        self._node_iterations[node.addr] += 1
-        self.type_constraints |= state.type_constraints
-        for var, typevar in state.typevars._typevars.items():
-            self.var_to_typevars[var].add(typevar)
+        self._node_iterations[block_key] += 1
 
         if state.ret_val_size is not None:
             if self.ret_val_size is None or self.ret_val_size < state.ret_val_size:
                 self.ret_val_size = state.ret_val_size
 
         state.downsize()
-        self._outstates[node.addr] = state
+        self._outstates[block_key] = state
 
         return True, state
 
@@ -433,6 +456,8 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
         pass
 
     def _post_analysis(self):
+        VariableRecoveryBase._post_analysis(self)
+
         self.variable_manager["global"].assign_variable_names(labels=self.kb.labels)
         self.variable_manager[self.function.addr].assign_variable_names()
 
@@ -444,15 +469,24 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
                     state.downsize_region(state.stack_region),
                 )
 
+        if self._unify_variables:
+            self.variable_manager[self.function.addr].unify_variables()
+
+        # fill in var_to_typevars
+        for var, typevar_set in self.typevars._typevars.items():
+            self.var_to_typevars[var] = typevar_set
+
         # unify type variables for global variables
         for var, typevars in self.var_to_typevars.items():
             if len(typevars) > 1 and isinstance(var, SimMemoryVariable) and not isinstance(var, SimStackVariable):
                 sorted_typevars = list(sorted(typevars, key=lambda x: str(x)))  # pylint:disable=unnecessary-lambda
                 for tv in sorted_typevars[1:]:
-                    self.type_constraints.add(Equivalence(sorted_typevars[0], tv))
+                    self.type_constraints[self.func_typevar].add(Equivalence(sorted_typevars[0], tv))
 
         self.variable_manager[self.function.addr].ret_val_size = self.ret_val_size
 
+        self.delayed_type_constraints = None
+
     #
     # Private methods
     #
