@@ -2,10 +2,9 @@ from .plugin import KnowledgeBasePlugin
 
 
 class Comments(KnowledgeBasePlugin, dict):
-
-    def __init__(self, kb):
-        super(Comments, self).__init__()
-        self._kb = kb
+    """
+    Tracks comments via a Dict of Address -> Text
+    """
 
     def copy(self):
         o = Comments(self._kb)
@@ -13,4 +12,4 @@ class Comments(KnowledgeBasePlugin, dict):
         return o
 
 
-KnowledgeBasePlugin.register_default('comments', Comments)
+KnowledgeBasePlugin.register_default("comments", Comments)
