@@ -1,7 +1,8 @@
 # pylint:disable=raise-missing-from
-from typing import Dict, Set, Optional
+from typing import Dict, Generator, Optional, Set
 import logging
 import collections.abc
+import re
 from sortedcontainers import SortedDict
 
 import networkx
@@ -14,6 +15,9 @@ from .function import Function
 from .soot_function import SootFunction
 
 
+QUERY_PATTERN = re.compile(r"^(::(.+?))?::(.+)$")
+ADDR_PATTERN = re.compile(r"^(0x[\dA-Fa-f]+)|(\d+)$")
+
 l = logging.getLogger(name=__name__)
 
 
@@ -76,8 +80,7 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
     """
 
     def __init__(self, kb):
-        super().__init__()
-        self._kb = kb
+        super().__init__(kb=kb)
         self.function_address_types = self._kb._project.arch.function_address_types
         self.address_types = self._kb._project.arch.address_types
         self._function_map: Dict[int, Function] = FunctionDict(self, key_types=self.function_address_types)
@@ -347,6 +350,11 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
     def get_by_addr(self, addr) -> Function:
         return self._function_map.get(addr)
 
+    def get_by_name(self, name: str) -> Generator[Function, None, None]:
+        for f in self._function_map.values():
+            if f.name == name:
+                yield f
+
     def _function_added(self, func: Function):
         """
         A callback method for adding a new function instance to the manager.
@@ -403,6 +411,37 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
         except KeyError:
             return None
 
+    def query(self, query: str) -> Optional[Function]:
+        """
+        Query for a function using selectors to disambiguate. Supported variations:
+
+            ::<name>           Function <name> in the main object
+            ::<addr>::<name>   Function <name> at <addr>
+            ::<obj>::<name>    Function <name> in <obj>
+
+        """
+        # FIXME: Proper mangle handling
+        matches = QUERY_PATTERN.match(query)
+        if matches:
+            selector = matches.group(2)
+            name = matches.group(3)
+
+            if selector is not None and ADDR_PATTERN.fullmatch(selector):
+                addr = int(matches.group(2), 0)
+                try:
+                    func = self._function_map.get(addr)
+                    if func.name == name:
+                        return func
+                except KeyError:
+                    pass
+
+            obj_name = selector or self._kb._project.loader.main_object.binary_basename
+            for func in self.get_by_name(name):
+                if func.binary_name == obj_name:
+                    return func
+
+        return None
+
     def function(self, addr=None, name=None, create=False, syscall=False, plt=None) -> Optional[Function]:
         """
         Get a function object from the function manager.
@@ -433,10 +472,13 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
                         f.is_syscall = True
                     return f
         elif name is not None:
-            for func in self._function_map.values():
-                if func.name == name:
-                    if plt is None or func.is_plt == plt:
-                        return func
+            func = self.query(name)
+            if func is not None:
+                return func
+
+            for func in self.get_by_name(name):
+                if plt is None or func.is_plt == plt:
+                    return func
 
         return None
 
@@ -447,6 +489,8 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
 
     def rebuild_callgraph(self):
         self.callgraph = networkx.MultiDiGraph()
+        for func_addr in self._function_map.keys():
+            self.callgraph.add_node(func_addr)
         for func in self._function_map.values():
             if func.block_addrs_set:
                 for node in func.transition_graph.nodes():
