@@ -1,4 +1,7 @@
+from typing import Optional
+
 import claripy
+
 from ..errors import SimMemoryError
 
 
@@ -24,6 +27,7 @@ class SimMemoryObject:
         "object",
         "length",
         "endness",
+        "_concrete_bytes",
     )
 
     def __init__(self, obj, base, endness, length=None, byte_width=8):
@@ -41,6 +45,7 @@ class SimMemoryObject:
         self.object = obj
         self.length = obj_bit_size(obj) // self._byte_width if length is None else length
         self.endness = endness
+        self._concrete_bytes: Optional[bytes] = None
 
     def size(self):
         return self.length * self._byte_width
@@ -64,6 +69,16 @@ class SimMemoryObject:
     def last_addr(self):
         return self.base + self.length - 1
 
+    def concrete_bytes(self, offset: int, size: int) -> Optional[bytes]:
+        if self._concrete_bytes is None:
+            if isinstance(self.object, claripy.ast.Bits) and self.object.op == "BVV" and not self.object.annotations:
+                self._concrete_bytes = self.object.concrete_value.to_bytes(len(self.object) // self._byte_width, "big")
+
+        if self._concrete_bytes is None:
+            return None
+
+        return self._concrete_bytes[offset : offset + size]
+
     def includes(self, x):
         return 0 <= x - self.base < self.length
 
@@ -148,7 +163,7 @@ class SimLabeledMemoryObject(SimMemoryObject):
         self.label = label
 
 
-def bv_slice(value, offset, size, rev, bw):
+def bv_slice(value: claripy.ast.BV, offset: int, size: int, rev: bool, bw: int) -> claripy.ast.BV:
     """
     Extremely cute utility to pretend you've serialized a value to stored bytes, sliced it a la python slicing, and then
     deserialized those bytes to an integer again.
@@ -176,4 +191,6 @@ def bv_slice(value, offset, size, rev, bw):
         return value
 
     bitstart = len(value) - offset * bw
+    if size == 0:
+        return claripy.BVV(b"")
     return value[bitstart - 1 : bitstart - size * bw]
