@@ -1,8 +1,9 @@
-from typing import Tuple, Optional, Dict, List
+from typing import Tuple, Optional, Dict, List, Set
 from collections import defaultdict
 import logging
 
 import networkx
+import networkx.algorithms
 
 
 def shallow_reverse(g) -> networkx.DiGraph:
@@ -66,9 +67,7 @@ def to_acyclic_graph(
 
     if ordered_nodes is None:
         # take the quasi-topological order of the graph
-        from angr.analyses.cfg.cfg_utils import CFGUtils  # pylint:disable=import-outside-toplevel
-
-        ordered_nodes = CFGUtils.quasi_topological_sort_nodes(graph, loop_heads=loop_heads)
+        ordered_nodes = GraphUtils.quasi_topological_sort_nodes(graph, loop_heads=loop_heads)
 
     acyclic_graph = networkx.DiGraph()
 
@@ -86,31 +85,33 @@ def to_acyclic_graph(
 
 def dfs_back_edges(graph, start_node):
     """
-    Do a DFS traversal of the graph, and return with the back edges.
-
-    Note: This is just a naive recursive implementation, feel free to replace it.
-    I couldn't find anything in networkx to do this functionality. Although the
-    name suggest it, but `dfs_labeled_edges` is doing something different.
+    Perform an iterative DFS traversal of the graph, returning back edges.
 
     :param graph:       The graph to traverse.
-    :param start_node:  The node where to start the traversal
-    :returns:           An iterator of 'backward' edges
+    :param start_node:  The node where to start the traversal.
+    :returns:           An iterator of 'backward' edges.
     """
+    if start_node not in graph:
+        return  # Ensures that the start node is in the graph
 
-    visited = set()
-    finished = set()
+    visited = set()  # Tracks visited nodes
+    finished = set()  # Tracks nodes whose descendants are fully explored
+    stack = [(start_node, iter(graph[start_node]))]
 
-    def _dfs_back_edges_core(node):
+    while stack:
+        node, children = stack[-1]
         visited.add(node)
-        for child in iter(graph[node]):
-            if child not in finished:
-                if child in visited:
-                    yield node, child
-                else:
-                    yield from _dfs_back_edges_core(child)
-        finished.add(node)
 
-    yield from _dfs_back_edges_core(start_node)
+        try:
+            child = next(children)
+            if child in visited:
+                if child not in finished:
+                    yield node, child  # Found a back edge
+            elif child not in finished:  # Check if the child has not been finished
+                stack.append((child, iter(graph[child])))
+        except StopIteration:
+            stack.pop()  # Done with this node's children
+            finished.add(node)  # Mark this node as finished
 
 
 def subgraph_between_nodes(graph, source, frontier, include_frontier=False):
@@ -146,6 +147,9 @@ def subgraph_between_nodes(graph, source, frontier, include_frontier=False):
         traversed.add(node)
 
         for _, succ, data in graph.out_edges(node, data=True):
+            if g0.has_edge(node, succ):
+                continue
+
             g0.add_edge(node, succ, **data)
             if succ in traversed or succ in frontier:
                 continue
@@ -295,6 +299,10 @@ class ContainerNode:
 
 
 class Dominators:
+    """
+    Describes dominators in a graph.
+    """
+
     dom: networkx.DiGraph
 
     def __init__(self, graph, entry_node, successors_func=None, reverse=False):
@@ -465,6 +473,8 @@ class Dominators:
             stack = [start_node]
             while len(stack) > 0:
                 node = stack.pop()
+                if node in scanned_nodes:
+                    continue
                 counter += 1
 
                 # Mark it as scanned
@@ -483,7 +493,6 @@ class Dominators:
                     if s not in scanned_nodes:
                         stack.append(s)
                         parent[s] = node
-                        scanned_nodes.add(s)
 
             if counter >= all_nodes_count:
                 break
@@ -527,9 +536,288 @@ class Dominators:
 
 
 class PostDominators(Dominators):
+    """
+    Describe post-dominators in a graph.
+    """
+
     def __init__(self, graph, entry_node, successors_func=None):
         super().__init__(graph, entry_node, successors_func=successors_func, reverse=True)
 
     @property
     def post_dom(self) -> networkx.DiGraph:
         return self.dom
+
+
+class SCCPlaceholder:
+    """
+    Describes a placeholder for strongly-connected-components in a graph.
+    """
+
+    __slots__ = ("scc_id",)
+
+    def __init__(self, scc_id):
+        self.scc_id = scc_id
+
+    def __eq__(self, other):
+        return isinstance(other, SCCPlaceholder) and other.scc_id == self.scc_id
+
+    def __hash__(self):
+        return hash("scc_placeholder_%d" % self.scc_id)
+
+
+class GraphUtils:
+    """
+    A helper class with some static methods and algorithms implemented, that in fact, might take more than just normal
+    CFGs.
+    """
+
+    @staticmethod
+    def find_merge_points(function_addr, function_endpoints, graph):  # pylint:disable=unused-argument
+        """
+        Given a local transition graph of a function, find all merge points inside, and then perform a
+        quasi-topological sort of those merge points.
+
+        A merge point might be one of the following cases:
+        - two or more paths come together, and ends at the same address.
+        - end of the current function
+
+        :param int function_addr: Address of the function.
+        :param list function_endpoints: Endpoints of the function. They typically come from Function.endpoints.
+        :param networkx.DiGraph graph: A local transition graph of a function. Normally it comes from Function.graph.
+        :return: A list of ordered addresses of merge points.
+        :rtype: list
+        """
+
+        merge_points = set()
+
+        for node in graph.nodes():
+            if graph.in_degree(node) > 1:
+                merge_points.add(node)
+
+        ordered_merge_points = GraphUtils.quasi_topological_sort_nodes(graph, merge_points)
+
+        addrs = [n.addr for n in ordered_merge_points]
+        return addrs
+
+    @staticmethod
+    def find_widening_points(function_addr, function_endpoints, graph):  # pylint: disable=unused-argument
+        """
+        Given a local transition graph of a function, find all widening points inside.
+
+        Correctly choosing widening points is very important in order to not lose too much information during static
+        analysis. We mainly consider merge points that has at least one loop back edges coming in as widening points.
+
+        :param int function_addr: Address of the function.
+        :param list function_endpoints: Endpoints of the function, typically coming from Function.endpoints.
+        :param networkx.DiGraph graph: A local transition graph of a function, normally Function.graph.
+        :return: A list of addresses of widening points.
+        :rtype: list
+        """
+
+        sccs = networkx.strongly_connected_components(graph)
+
+        widening_addrs = set()
+
+        for scc in sccs:
+            if len(scc) == 1:
+                node = next(iter(scc))
+                if graph.has_edge(node, node):
+                    # self loop
+                    widening_addrs.add(node.addr)
+            else:
+                for n in scc:
+                    predecessors = graph.predecessors(n)
+                    if any(p not in scc for p in predecessors):
+                        widening_addrs.add(n.addr)
+                        break
+
+        return list(widening_addrs)
+
+    @staticmethod
+    def reverse_post_order_sort_nodes(graph, nodes=None):
+        """
+        Sort a given set of nodes in reverse post ordering.
+
+        :param networkx.DiGraph graph: A local transition graph of a function.
+        :param iterable nodes: A collection of nodes to sort.
+        :return: A list of sorted nodes.
+        :rtype: list
+        """
+
+        post_order = networkx.dfs_postorder_nodes(graph)
+
+        if nodes is None:
+            return reversed(list(post_order))
+
+        addrs_to_index = {n.addr: i for (i, n) in enumerate(post_order)}
+        return sorted(nodes, key=lambda n: addrs_to_index[n.addr], reverse=True)
+
+    @staticmethod
+    def quasi_topological_sort_nodes(
+        graph: networkx.DiGraph, nodes: Optional[List] = None, loop_heads: Optional[List] = None
+    ) -> List:
+        """
+        Sort a given set of nodes from a graph based on the following rules:
+
+        # - if A -> B and not B -> A, then we have A < B
+        # - if A -> B and B -> A, then the ordering is undefined
+
+        Following the above rules gives us a quasi-topological sorting of nodes in the graph. It also works for cyclic
+        graphs.
+
+        :param graph:       A local transition graph of the function.
+        :param nodes:       A list of nodes to sort. None if you want to sort all nodes inside the graph.
+        :param loop_heads:  A list of nodes that should be treated loop heads.
+        :return:            A list of ordered nodes.
+        """
+
+        # fast path for single node graphs
+        if graph.number_of_nodes() == 1:
+            if nodes is None:
+                return list(graph.nodes)
+            return [n for n in graph.nodes() if n in nodes]
+
+        # make a copy to the graph since we are gonna modify it
+        graph_copy = networkx.DiGraph()
+
+        # find all strongly connected components in the graph
+        sccs = [scc for scc in networkx.strongly_connected_components(graph) if len(scc) > 1]
+
+        def _sort_edge(edge):
+            """
+            A sorter to make a deterministic order of edges.
+            """
+            _src, _dst = edge
+            src_addr, dst_addr = 0, 0
+            if hasattr(_src, "addr"):
+                src_addr = _src.addr
+            elif isinstance(_src, int):
+                src_addr = _src
+
+            if hasattr(_dst, "addr"):
+                dst_addr = _dst.addr
+            elif isinstance(_dst, int):
+                dst_addr = _dst
+
+            return src_addr + dst_addr
+
+        # collapse all strongly connected components
+        edges = sorted(list(graph.edges()), key=_sort_edge)
+        for src, dst in edges:
+            scc_index = GraphUtils._components_index_node(sccs, src)
+            if scc_index is not None:
+                src = SCCPlaceholder(scc_index)
+            scc_index = GraphUtils._components_index_node(sccs, dst)
+            if scc_index is not None:
+                dst = SCCPlaceholder(scc_index)
+
+            if isinstance(src, SCCPlaceholder) and isinstance(dst, SCCPlaceholder) and src == dst:
+                if src not in graph_copy:
+                    graph_copy.add_node(src)
+                continue
+            if src == dst:
+                if src not in graph_copy:
+                    graph_copy.add_node(src)
+                continue
+
+            graph_copy.add_edge(src, dst)
+
+        # add loners
+        out_degree_zero_nodes = [node for (node, degree) in graph.out_degree() if degree == 0]
+        for node in out_degree_zero_nodes:
+            if graph.in_degree(node) == 0:
+                graph_copy.add_node(node)
+
+        # topological sort on acyclic graph `graph_copy`
+        tmp_nodes = networkx.topological_sort(graph_copy)
+
+        ordered_nodes = []
+        for n in tmp_nodes:
+            if isinstance(n, SCCPlaceholder):
+                GraphUtils._append_scc(graph, ordered_nodes, sccs[n.scc_id], loop_head_candidates=loop_heads)
+            else:
+                ordered_nodes.append(n)
+
+        if nodes is None:
+            return ordered_nodes
+
+        nodes = set(nodes)
+        ordered_nodes = [n for n in ordered_nodes if n in nodes]
+        return ordered_nodes
+
+    @staticmethod
+    def _components_index_node(components, node):
+        for i, comp in enumerate(components):
+            if node in comp:
+                return i
+        return None
+
+    @staticmethod
+    def _append_scc(
+        graph: networkx.DiGraph, ordered_nodes: List, scc: Set, loop_head_candidates: Optional[List] = None
+    ) -> None:
+        """
+        Append all nodes from a strongly connected component to a list of ordered nodes and ensure the topological
+        order.
+
+        :param graph: The graph where all nodes belong to.
+        :param ordered_nodes:     Ordered nodes.
+        :param scc:           A set of nodes that forms a strongly connected component in the graph.
+        """
+
+        loop_head = None
+
+        if loop_head_candidates is not None:
+            # find the first node that appears in loop_heads
+            loop_head_candidates = set(loop_head_candidates)
+            for n in scc:
+                if n in loop_head_candidates:
+                    loop_head = n
+                    break
+
+        if loop_head is None:
+            for parent_node in reversed(ordered_nodes):
+                # find all successors to this node
+                succs = set(graph.successors(parent_node))
+                scc_succs = scc.intersection(succs)
+                if len(scc_succs) == 1:
+                    loop_head = next(iter(scc_succs))
+                    break
+                if len(scc_succs) > 1:
+                    # calculate the distance between each pair of nodes within scc_succs, pick the one with the
+                    # shortest total distance
+                    scc_node_distance = defaultdict(int)
+                    for scc_succ in scc_succs:
+                        for other_node in scc_succs:
+                            if other_node is scc_succ:
+                                continue
+                            scc_node_distance[scc_succ] += networkx.algorithms.shortest_path_length(
+                                graph, scc_succ, other_node
+                            )
+                    distance_to_node = {v: k for k, v in scc_node_distance.items()}
+                    lowest_distance = min(distance_to_node)
+                    loop_head = distance_to_node[lowest_distance]
+                    break
+
+        if loop_head is None:
+            # randomly pick one
+            loop_head = next(iter(scc))
+
+        subgraph: networkx.DiGraph = graph.subgraph(scc).copy()
+        for src, _ in list(subgraph.in_edges(loop_head)):
+            subgraph.remove_edge(src, loop_head)
+
+        # panic mode: if the strongly connected component has too many edges (imagine an almost complete graph), it
+        # will take too long to converge if we only remove one node out of the component each time. we introduce a
+        # panic mode that will aggressively remove edges
+
+        if len(subgraph) > 3000 and len(subgraph.edges) > len(subgraph) * 1.4:
+            for n in scc:
+                if subgraph.in_degree[n] >= 1 and subgraph.out_degree[n] >= 1:
+                    for src in list(subgraph.predecessors(n)):
+                        if src is not n:
+                            subgraph.remove_edge(src, n)
+                            if len(subgraph.edges) <= len(subgraph) * 1.4:
+                                break
+
+        ordered_nodes.extend(GraphUtils.quasi_topological_sort_nodes(subgraph))
